import io
import copy
import json
import numpy as np
import ase
import itertools

from collections import namedtuple
from collections.abc import Callable
from typing import Self

from mendeleev import element
from ase.optimize import FIRE

from rdkit import Chem
from rdkit.Chem import rdMolTransforms, AllChem, rdMolAlign, rdmolops
from rdkit.Chem.Draw import rdMolDraw2D
from PIL import Image

from rdworks.std import clean_2d
from rdworks.torsion import get_torsion_atoms, create_torsion_fragment
from rdworks.units import ev2kcalpermol, pm2angstrom
from rdworks.utils import recursive_round, compress_string, decompress_string
from rdworks.xtb.wrapper import GFN2xTB
from rdworks.display import render_png, render_svg


class Conf:
    """Container for 3D conformers."""
    
    def __init__(self, 
                 molecule: str | Chem.Mol | None = None, 
                 name: str = '',
                 compressed: bool = False) -> None:
        """Initialize.

        Args:
            molecule (Chem.Mol | MolBlock string): Molecule for 3D conformer.
            name (str): Name prefix of the generated conformers. Defaults to ''.
            compressed (bool): whether the MolBlock string is compressed or not. 
                Defaults to False.

        Raises:
            ValueError: if `molecule` is not rdkit.Chem.Mol object.
        """
        assert isinstance(molecule, str | Chem.Mol) or molecule is None

        self.name = name
        self.rdmol = None   # must contain one and only one rdkit conformer
        self.natoms = 0
        self.charge = 0     # molecular formal charge
        self.spin = 1       # molecular spin multiplicity for ASE
        # Molecular spin multiplicity describes the number of possible orientations of 
        # spin angular momentum for a given molecule, essentially indicating the total 
        # number of unpaired electrons.
        # Spin Angular Momentum (S): up (+1/2) or down (-1/2)
        # Spin Multiplicity (2S + 1)
        # 0 unpaired electron  has S = 0,   2S + 1 = 0, called a singlet. 
        # 1 unpaired electron  has S = 1/2, 2S + 1 = 2, called a doublet (radical). 
        # 2 unpaired electrons has S = 1,   2S + 1 = 3, called a triplet.  
        
        self.props = {}

        if molecule is None:
            return

        if isinstance(molecule, str): # 3-D MolBLock string
            if compressed:
                molecule = decompress_string(molecule)
            try:
                self.rdmol = Chem.MolFromMolBlock(molecule, 
                                                  sanitize=False, 
                                                  removeHs=False, 
                                                  strictParsing=True)
            except:
                ValueError(f'Conf() Error: invalid MolBlock string')

        elif isinstance(molecule, Chem.Mol): # 3-D
            try:
                self.rdmol = molecule
            except:
                ValueError(f'Conf() Error: invalid Chem.Mol object')

        num_atoms = self.rdmol.GetNumAtoms()
        tot_atoms = self.rdmol.GetNumAtoms(onlyExplicit=False)
        
        assert num_atoms == tot_atoms, "Conf() Error: missing hydrogens"
        
        self.natoms = num_atoms
        self.charge = rdmolops.GetFormalCharge(self.rdmol)
        self.props.update({'atoms': self.natoms, 'charge': self.charge})
     
        assert self.rdmol.GetConformer().Is3D(), "Conf() Error: not 3D"

        

    def __str__(self) -> str:
        """Returns a string representation.

        Returns:
            str: string representation.
        """
        return f"<rdworks.Conf({self.rdmol} name={self.name} atoms={self.natoms})>"
       

    def copy(self) -> Self:
        """Returns a copy of self.

        Returns:
            Self: `rdworks.Conf` object.
        """
        return copy.deepcopy(self)
    

    def rename(self, name:str) -> Self:
        """Rename and returns self.

        Args:
            name (str): a new name for conformers.

        Raises:
            ValueError: if `name` is not given.

        Returns:
            Self: `rdworks.Conf` object.
        """
        if not name:
            raise ValueError('rdworks.Conf.rename() expects a name')
        self.name = name
        self.rdmol.SetProp('_Name', name)
        return self
    

    def sync(self, coord: np.ndarray | list) -> Self:
        """Synchronize the conformer coordinates with the provided `coord`.

        Args:
            coord (np.array): 3D coordinates.

        Raises:
            ValueError: if `coord` does not have the correct shape (natoms, 3).

        Returns:
            Self: `rdworks.Conf` object.
        """
        if isinstance(coord, np.ndarray) and coord.shape != (self.natoms, 3):
            raise ValueError(f"`coord.shape` should be ({self.natoms},3)")
        elif isinstance(coord, list) and len(coord) != self.natoms:
            raise ValueError(f"`coord` should be length of {self.natoms}")
        for i, a in enumerate(self.rdmol.GetAtoms()):
            self.rdmol.GetConformer().SetAtomPosition(a.GetIdx(), coord[i])
        
        return self


    def set_torsion(self, i:int, j:int, k:int, l:int, degree:float) -> Self:
        """Set dihedral angle (i-j-k-l) in degrees.

        Args:
            i (int): atom index
            j (int): atom index
            k (int): atom index
            l (int): atom index
            degree (float): dihedral angle in degrees

        Returns:
            Self: modified Conf object
        """
        rdMolTransforms.SetDihedralDeg(self.rdmol.GetConformer(), i, j, k, l, degree)

        return self
    
    
    def optimize(self, 
                 calculator: str | Callable = 'MMFF94',
                 fmax:float=0.05, 
                 max_iter:int=1000,
                 **kwargs) -> Self:
        """Optimize 3D geometry.

        Args:
            calculator (str | Callable): MMFF94 (= MMFF), MMFF94s, UFF, or ASE calculator.
                `MMFF94` or `MMFF` - Intended for general use, including organic molecules and proteins, 
                    and primarily relies on data from quantum mechanical calculations. 
                    It's often used in molecular dynamics simulations.
                `MMFF94s` - A "static" variant of MMFF94, with adjusted parameters for out-of-plane 
                    bending and dihedral torsions to favor planar geometries for specific nitrogen atoms. 
                    This makes it better suited for geometry optimization studies where a static, 
                    time-averaged structure is desired. The "s" stands for "static".
                `UFF` - UFF refers to the "Universal Force Field," a force field model used for 
                    molecular mechanics calculations. It's a tool for geometry optimization, 
                    energy minimization, and exploring molecular conformations in 3D space. 
                    UFF is often used to refine conformers generated by other methods, 
                    such as random conformer generation, to produce more physically plausible 
                    and stable structures.
            fmax (float, optional): fmax for the calculator convergence. Defaults to 0.05.
            max_iter (int, optional): max iterations for the calculator. Defaults to 1000.
            
        Args for xTB:
            water (str, optional): water solvation model (choose 'gbsa' or 'alpb')
                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
                gbsa: generalized Born (GB) model with Surface Area contributions.

        Returns:
            Self: self
        """
        if isinstance(calculator, str) :
            
            PE_start = self.singlepoint(calculator)
            PE_final = None
            
            if calculator.lower() == 'xTB'.lower():
                water = kwargs.get('water', None)
                results = GFN2xTB(self.rdmol).optimize(water=water)
                # SimpleNamespace(
                #         PE = datadict['total energy'] * hartree2kcalpermol,
                #         charges = datadict['partial charges'],
                #         wbo = Wiberg_bond_orders,
                #         geometry = rdmol_opt,
                # )
                try:
                    self.rdmol = results.geometry
                    PE_final = results.PE
                    retcode = 0
                except:
                    retcode = 1
                
            elif calculator.lower() == 'MMFF94'.lower() or calculator.lower() == 'MMFF'.lower():
                retcode = Chem.rdForceFieldHelpers.MMFFOptimizeMolecule(self.rdmol, 
                                                                        mmffVariant='MMFF94',
                                                                        maxIters=max_iter)
                # returns 0 if the optimization converged
            elif calculator.lower() == 'MMFF94s'.lower():
                retcode = Chem.rdForceFieldHelpers.MMFFOptimizeMolecule(self.rdmol, 
                                                                        mmffVariant='MMFF94s',
                                                                        maxIters=max_iter)
                # returns 0 if the optimization converged
            elif calculator.lower() == 'UFF'.lower():
                retcode = Chem.rdForceFieldHelpers.UFFOptimizeMolecule(self.rdmol,
                                                                       maxIters=max_iter)
                # returns 0 if the optimization converged

            if PE_final is None:
                PE_final = self.singlepoint(calculator)
            
            self.props.update({
                'E_tot_init(kcal/mol)': PE_start , # energy before optimization
                'E_tot(kcal/mol)': PE_final, # energy after optimization
                'Converged' : retcode == 0, # True or False
            })
            
            return self

        else:
            # assuming ASE calculator
            with io.StringIO() as logfile:
                ase_atoms = ase.Atoms(symbols=self.symbols(), positions=self.positions())
                ase_atoms.info['charge'] = self.charge
                ase_atoms.info['spin'] = self.spin
                ase_atoms.calc = calculator
                FIRE(ase_atoms, logfile=logfile).run(fmax=fmax)
                lines = [l.strip().split()[1:] for l in logfile.getvalue().split('\n') if l.startswith('FIRE')]
                data = [(float(e), float(f)) for (_, _, e, f) in lines]
                self.props.update({
                    'E_tot_init(kcal/mol)': data[0][0] * ev2kcalpermol, # energy before optimization
                    'E_tot(kcal/mol)': data[-1][0] * ev2kcalpermol, # energy after optimization
                    'Converged' : data[-1][1] < fmax, # True or False
                })
                
                # update atomic coordinates
                return self.sync(ase_atoms.get_positions())
 
    
    def protonate(self, atom_indices: list[int]) -> Self:
        """Protonate given non-hydrogen atoms.

        Args:
            atom_indices (list[int]): atom indices of non-hydrogen atoms to protonate.

        Returns:
            Self: self.
        """
        for idx in atom_indices:
            atom = self.rdmol.GetAtomWithIdx(idx)
            h = atom.GetNumExplicitHs()
            c = atom.GetFormalCharge()
            atom.SetNumExplicitHs(h+1)
            atom.SetFormalCharge(c+1)            
            Chem.SanitizeMol(self.rdmol)
            self.rdmol = Chem.AddHs(self.rdmol, addCoords=True)
            # The Chem.AddHs function in RDKit returns a new Mol object with hydrogens added to the molecule. 
            # It modifies the input molecule by adding hydrogens, 
            # but the original molecule remains unchanged.
        
        return self
    

    def deprotonate(self, atom_indices: list[int]) -> Self:
        """Deprotonate given non-hydrogen atoms.

        Args:
            atom_indices (list[int]): atom indices of non-hydrogen atoms to deprotonate.

        Returns:
            Self: self.
        """
        for idx in atom_indices:
            bonded_H_idx = None
            atom = self.rdmol.GetAtomWithIdx(idx)
            h = atom.GetNumExplicitHs()
            if h-1 >= 0 :
                atom.SetNumExplicitHs(h-1) # (h-1) must be unsigned int
            c = atom.GetFormalCharge()
            atom.SetFormalCharge(c-1)            
            neighbors = atom.GetNeighbors()
            
            for neighbor in neighbors:
                if neighbor.GetAtomicNum() == 1:
                    bonded_H_idx = neighbor.GetIdx()
                    break

            if bonded_H_idx is not None:
                edit_mol = Chem.EditableMol(self.rdmol)
                edit_mol.RemoveAtom(bonded_H_idx)
                self.rdmol = edit_mol.GetMol()
                Chem.SanitizeMol(self.rdmol)
        
        return self


    ##################################################
    ### Endpoint methods
    ##################################################

    def check_bonds(self, tolerance:float=0.25) -> bool:
        """Check bond lengths.

        Args:
            tolerance (float, optional): tolerance from the sum of 
                van der Waals radii of bonded atoms. Defaults to 0.25 (A).

        Returns:
            bool: True if all bond lengths are accceptable.
        """

        for bond in self.rdmol.GetBonds():
            idx1 = bond.GetBeginAtomIdx()
            idx2 = bond.GetEndAtomIdx()
            nuc1 = self.rdmol.GetAtomWithIdx(idx1).GetAtomicNum()
            nuc2 = self.rdmol.GetAtomWithIdx(idx2).GetAtomicNum()
            sum_radii = (element(nuc1).vdw_radius + element(nuc2).vdw_radius) * pm2angstrom
            bond_length = rdMolTransforms.GetBondLength(self.rdmol.GetConformer(), idx1, idx2)
            if abs(bond_length - sum_radii) > tolerance:
                return False
            
        return True
    

    def charge(self) -> int:
        """Returns molecular formal charge.

        Returns:
            int: charge
        """
        return self.charge
    

    def positions(self) -> np.array:
        """Returns the coordinates.

        Returns:
            np.array: the coordinates.
        """
        return np.array(self.rdmol.GetConformer().GetPositions().tolist())


    def symbols(self) -> list[str]:
        """Returns the element symbols.

        Returns:
            list: list of element symbols.
        """
        return [ a.GetSymbol() for a in self.rdmol.GetAtoms() ]


    def numbers(self) -> list[int]:
        """Returns the atomic numbers.

        Returns:
            list: list of atomic numbers.
        """
        return [ a.GetAtomicNum() for a in self.rdmol.GetAtoms() ]


    def cog(self) -> np.array:
        """Returns the center of geometry (COG).

        Returns:
            np.array: the center of geometry (COG).
        """
        xyz = []
        for i in range(0, self.natoms):
            pos = self.rdmol.GetConformer().GetAtomPositions(i)
            xyz.append([pos.x, pos.y, pos.z])
        return np.mean(xyz, axis=0)


    def rg(self) -> float:
        """Returns the radius of gyration (Rg).

        Returns:
            float: the radius of gyration (Rg).
        """
        xyz = []
        for i in range(0, self.natoms):
            pos = self.rdmol.GetConformer().GetAtomPositions(i)
            xyz.append([pos.x, pos.y, pos.z])
        xyz = np.array(xyz)
        cog = np.mean(xyz, axis=0)
        a = xyz-cog
        b = np.einsum('ij,ij->i', a, a)
        return np.sqrt(np.mean(b))


    def singlepoint(self, calculator: str | Callable = 'MMFF94', **kwargs) -> float:
        """Get potential energy and set `E_tot(kcal/mol)` in the self.props.

        Args:
            calculator (str | Callable): MMFF94 (= MMFF), MMFF94s, UFF, or ASE calculator.
                `MMFF94` or `MMFF` - Intended for general use, including organic molecules and proteins, 
                    and primarily relies on data from quantum mechanical calculations. 
                    It's often used in molecular dynamics simulations.
                `MMFF94s` - A "static" variant of MMFF94, with adjusted parameters for out-of-plane 
                    bending and dihedral torsions to favor planar geometries for specific nitrogen atoms. 
                    This makes it better suited for geometry optimization studies where a static, 
                    time-averaged structure is desired. The "s" stands for "static".
                `UFF` - UFF refers to the "Universal Force Field," a force field model used for 
                    molecular mechanics calculations. It's a tool for geometry optimization, 
                    energy minimization, and exploring molecular conformations in 3D space. 
                    UFF is often used to refine conformers generated by other methods, 
                    such as random conformer generation, to produce more physically plausible 
                    and stable structures.

        Returns:
            float | None: potential energy in kcal/mol or None.
        """
        if isinstance(calculator, str):
            if calculator.lower() == 'xTB'.lower():
                water = kwargs.get('water', None)
                results = GFN2xTB(self.rdmol).singlepoint(water=water)
                # SimpleNamespace(
                #     PE = datadict['total energy'] * hartree2kcalpermol,
                #     Gsolv = Gsolv, 
                #     charges = datadict['partial charges'], 
                #     wbo = Wiberg_bond_orders,
                #     )
                PE = results.PE
            
            elif calculator.lower() == 'MMFF94'.lower() or calculator.lower() == 'MMFF'.lower():
                mp = Chem.rdForceFieldHelpers.MMFFGetMoleculeProperties(self.rdmol, mmffVariant='MMFF94')
                ff = Chem.rdForceFieldHelpers.MMFFGetMoleculeForceField(self.rdmol, mp)
                PE = ff.CalcEnergy()
            
            elif calculator.lower() == 'MMFF94s'.lower():
                mp = Chem.rdForceFieldHelpers.MMFFGetMoleculeProperties(self.rdmol, mmffVariant='MMFF94s')
                ff = Chem.rdForceFieldHelpers.MMFFGetMoleculeForceField(self.rdmol, mp)
                PE = ff.CalcEnergy()
            
            elif calculator.lower() == 'UFF'.lower():
                ff = Chem.rdForceFieldHelpers.UFFGetMoleculeForceField(self.rdmol)
                PE = ff.CalcEnergy()
            
            else:
                raise ValueError("Unsupported calculator")                
            
            self.props.update({'E_tot(kcal/mol)': PE})

            return PE
        
        else:
            try:
                ase_atoms = ase.Atoms(symbols=self.symbols(), positions=self.positions())
                ase_atoms.info['charge'] = self.charge
                ase_atoms.info['spin'] = self.spin
                ase_atoms.calc = calculator
                PE = ase_atoms.get_potential_energy() # np.array
                if isinstance(PE, float):
                    PE = ev2kcalpermol * PE
                elif isinstance(PE, np.ndarray | list):
                    PE = ev2kcalpermol * float(PE[0]) # np.float64 to float
                self.props.update({'E_tot(kcal/mol)': PE})
            
                return PE
            
            except:
                raise RuntimeError("ASE calculator error")


    def torsion_atoms(self, strict: bool = True) -> dict[int, tuple]:
        """Determine torsion/dihedral angle atoms (i-j-k-l) and rotating group for each rotatable bond (j-k).

        Args:
            strict (bool): whether to exclude amide/imide/ester/acid bonds.

        Returns:
            {torsion_key: (i, j, k, l), ...,}
        """
        return {i: d[:4] for i, d in enumerate(get_torsion_atoms(self.rdmol, strict))}
    

    def torsion_energies_one(self,
                             calculator: str | Callable,
                             indices: tuple,
                             simplify: bool = True,
                             fmax: float = 0.05,
                             interval: float = 20.0,
                             use_converged_only: bool = True,
                             **kwargs) -> Self:
        """Calculate potential energy profile for a torsion angle.

        Args:
            calculator (str | Callable): 'MMFF', 'UFF', or ASE calculator.
            indices (tuple): atom indices (i,j,k,l) for a torsion angle
            simplify (bool, optional): whether to use fragementation. Defaults to True.
            fmax (float, optional): convergence limit for optimize. Defaults to 0.05.
            interval (float, optional): angle intervals. Defaults to 20.0.
            use_converged_only (bool, optional): whether to use only converged data. Defaults to True.

        Returns:
            Self: modified self.
        """
        ref_conf = self.copy()

        data = {'indices': indices,
                'angle': [], 
                'init': [], 
                'last': [], 
                'Converged': [],
                }
        
        if simplify:
            (frag, frag_ijkl, frag_created, wbo_filtered) = create_torsion_fragment(ref_conf.rdmol, indices)
            frag_conf = Conf(frag)
            for angle in np.arange(-180.0, 180.0, interval): 
                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
                conf = frag_conf.copy()
                conf.set_torsion(*frag_ijkl, angle) # atoms bonded to `l` move.
                conf = conf.optimize(calculator, fmax, **kwargs)
                data['angle'].append(angle)
                # conf.optimize() updates coordinates and conf.props: 
                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
                data['init'].append(conf.props['E_tot_init(kcal/mol)'])
                data['last'].append(conf.props['E_tot(kcal/mol)'])
                data['Converged'].append(conf.props['Converged'])
                frag_cleaned, _ = clean_2d(frag, reset_isotope=True, remove_H=True)
                # to serialize the molecule
                data['frag'] = Chem.MolToMolBlock(frag_cleaned)
                data['frag_indices'] = frag_ijkl
        else:
            for angle in np.arange(-180.0, 180.0, interval): 
                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
                conf = ref_conf.copy()
                conf.set_torsion(*indices, angle) # atoms bonded to `l` move.
                conf = conf.optimize(calculator, fmax, **kwargs)
                data['angle'].append(angle)
                # conf.optimize() updates coordinates and conf.props: 
                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
                data['init'].append(conf.props['E_tot_init(kcal/mol)'])
                data['last'].append(conf.props['E_tot(kcal/mol)'])
                data['Converged'].append(conf.props['Converged'])
        
        # Post-processing        
        if use_converged_only:
            data['angle'] = list(itertools.compress(data['angle'], data['Converged']))
            data['init' ] = list(itertools.compress(data['init' ], data['Converged']))
            data['last' ] = list(itertools.compress(data['last' ], data['Converged']))
        
        relax = np.array(data['init']) - np.median(data['last'])
        E_rel = relax - np.min(relax)

        torsion_energy_profile = {
            'indices' : data['indices'],
            'angle' : np.round(np.array(data['angle']), 1).tolist(), # np.ndarray -> list for serialization
            'E_rel(kcal/mol)': np.round(E_rel, 2).tolist(), # np.ndarray -> list for serialization
            }
        
        if simplify:
            torsion_energy_profile.update({
                'frag' : data.get('frag', None),
                'frag_indices' : data.get('frag_indices', None),
            })

        self.props['torsion'] = torsion_energy_profile
        self.props['torsion_calculator'] = str(calculator)

        return self
    

    def torsion_energies_batch(self,
                             batchoptimizer: Callable,
                             indices: tuple,
                             simplify: bool = True,
                             fmax: float = 0.05,
                             interval: float = 20.0,
                             use_converged_only: bool = True,
                             **kwargs) -> Self:
        """Calculate potential energy profile for torsion angles in a batch.

        Args:
            batchoptimizer (Callable): BatchOptimizer.
            indices (tuple): atom indices (i,j,k,l) for a torsion angle
            simplify (bool, optional): whether to use fragementation. Defaults to True.
            fmax (float, optional): convergence limit for optimize. Defaults to 0.05.
            interval (float, optional): angle intervals. Defaults to 20.0.
            use_converged_only (bool, optional): whether to use only converged data. Defaults to True.

        Returns:
            Self: modified self.
        """
        ref_conf = self.copy()

        data = {'indices': indices,
                'angle': [], 
                'init': [], 
                'last': [], 
                'Converged': [],
                }
        
        confs = []
        if simplify:
            (frag, frag_ijkl, frag_created, wbo_filtered) = create_torsion_fragment(ref_conf.rdmol, indices)
            frag_conf = Conf(frag)
            for angle in np.arange(-180.0, 180.0, interval): 
                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
                conf = frag_conf.copy()
                conf.set_torsion(*frag_ijkl, angle) # atoms bonded to `l` move.
                frag_cleaned, _ = clean_2d(frag, reset_isotope=True, remove_H=True)
                # to serialize the molecule
                data['frag'] = Chem.MolToMolBlock(frag_cleaned)
                data['frag_indices'] = frag_ijkl
                data['angle'].append(angle)
                confs.append(conf)
      
        else:
            for angle in np.arange(-180.0, 180.0, interval): 
                # Iterated numpy.ndarray does not contain the last 180: -180., ..., (180).
                conf = ref_conf.copy()
                conf.set_torsion(*indices, angle) # atoms bonded to `l` move.
                data['angle'].append(angle)
                confs.append(conf)
                
        batchsize_atoms = kwargs.get('batchsize_atoms', 16384)
        batches = prep_confs_batches(confs, batchsize_atoms)

        for batch in batches:
            optimized = batchoptimizer(batch.rdmols).run()
            for rdmol in optimized.mols:
                # conf.optimize() updates coordinates and conf.props: 
                #   `E_tot_init(kcal/mol)`, `E_tot(kcal/mol)`, `Converged`.
                data['init'].append(float(rdmol.GetProp('E_tot_init(kcal/mol)')))
                data['last'].append(float(rdmol.GetProp('E_tot(kcal/mol)')))
                data['Converged'].append(True if rdmol.GetProp('Converged') == 'True' else False)

        # Post-processing        
        if use_converged_only:
            data['angle'] = list(itertools.compress(data['angle'], data['Converged']))
            data['init' ] = list(itertools.compress(data['init' ], data['Converged']))
            data['last' ] = list(itertools.compress(data['last' ], data['Converged']))
        
        relax = np.array(data['init']) - np.median(data['last'])
        E_rel = relax - np.min(relax)

        torsion_energy_profile = {
            'indices' : data['indices'],
            'angle' : np.round(np.array(data['angle']), 1).tolist(), # np.ndarray -> list for serialization
            'E_rel(kcal/mol)': np.round(E_rel, 2).tolist(), # np.ndarray -> list for serialization
            }
        
        if simplify:
            torsion_energy_profile.update({
                'frag' : data.get('frag', None),
                'frag_indices' : data.get('frag_indices', None),
            })

        self.props['torsion'] = torsion_energy_profile
        self.props['torsion_calculator'] = str(batchoptimizer)

        return self

        
    def torsion_energies(self,
                         calculator: str | Callable,
                         torsion_key: int | None = None,
                         simplify: bool = True, 
                         fmax: float = 0.05,
                         interval: float = 20.0,
                         use_converged_only: bool = True,
                         **kwargs,
                         ) -> Self:
        """Calculates potential energy profiles for each torsion angle using ASE optimizer.

        It uses the first conformer as a reference.

        Args:
            calculator (str | Callable): 'MMFF', 'UFF', or ASE calculator.
            torsion_key (int | None): key to the torsion indices to calculate. Defaults to None (all).
            simplify (bool, optional): whether to use fragment surrogate. Defaults to True.
            fmax (float, optional): fmax of ASE optimizer. Defaults to 0.05.
            interval (float, optional): interval of torsion angles in degree. Defaults to 15.0.
            use_converged_only (bool, optional): whether to use only converged data. Defaults to True.
            batchsize_atoms (int, optional): max number of atoms in a batch. 
                Defaults to 16384(=16*1024). Disable batch optimization if smaller than conf.natoms.

        Returns:
            Self: modified self.
        """

        torsion_atoms_dict = self.torsion_atoms() # {0: (5, 4, 3, 1)}
        if (torsion_key is not None) and torsion_atoms_dict.get(torsion_key):
            torsion_atoms_dict = {torsion_key: torsion_atoms_dict.get(torsion_key)}
            # single torsion angle atom indices

        conf = self.copy()
        torsion_energy_profiles = {}
        if 'batchsize_atoms' in kwargs and kwargs.get('batchsize_atoms') >= conf.natoms:
            batchsize_atoms = kwargs.get('batchsize_atoms')
            for tk, indices in torsion_atoms_dict.items():
                conf = conf.torsion_energies_batch(calculator, 
                                                indices, 
                                                simplify, 
                                                fmax, 
                                                interval, 
                                                use_converged_only,
                                                batchsize_atoms=batchsize_atoms,
                                                )
                torsion_energy_profiles[tk] = conf.props['torsion']
        else:
            for tk, indices in torsion_atoms_dict.items():
                conf = conf.torsion_energies_one(calculator, 
                                                indices, 
                                                simplify, 
                                                fmax,
                                                interval, 
                                                use_converged_only,
                                                )
                torsion_energy_profiles[tk] = conf.props['torsion']
        
        self.props['torsion'] = torsion_energy_profiles
        self.props['torsion_calculator'] = str(calculator)

        return self
    

    def torsion_angle(self, i: int, j: int, k: int, l: int) -> float:
        """Get dihedral angle (i-j-k-l) in degrees.

        Args:
            i (int): atom index
            j (int): atom index
            k (int): atom index
            l (int): atom index

        Returns:
            float: dihedral angle in degrees.
        """
        degree = rdMolTransforms.GetDihedralDeg(self.rdmol.GetConformer(), i, j, k, l)

        return degree


    def dumps(self, key:str='') -> str:
        """Returns JSON dumps of the `props`.

        Args:
            key (str): a key for the `props` dictionary. Defaults to '' (all).

        Returns:
            str: JSON dumps.
        """
        if key:
            return json.dumps({key:self.props[key]})
        else:
            return json.dumps(self.props)
    

    def serialize(self, decimals: int = 3, compressed: bool = False) -> str:
        """Serialize information necessary to rebuild a Conf object.
        Args:
            decimals (int, optional): number of decimal places for float data type. Defaults to 3.

        Returns:
            str: serialized string for json.loads()
        """
        serialized = json.dumps({
            'name' : self.name,
            'natoms': self.natoms,
            'charge': self.charge,
            'spin': self.spin,
            'props' : recursive_round(self.props, decimals),
            'molblock' : self.to_molblock(),
            })

        if compressed:
            serialized = compress_string(serialized)
            
        return serialized


    def deserialize(self, serialized: str, compressed: bool = False) -> Self:
        """De-serialize information and rebuild a Conf object.
        
        Example:
            serialized = conf1.serialize()
            conf2 = Conf().deserialize(serialized)

        Args:
            serialized (str): serialized string.

        Returns:
            Self: modified self.
        """
        if compressed:
            serialized = decompress_string(serialized)

        data = json.loads(serialized)

        self.name = data['name']
        self.natoms = int(data['natoms'])
        self.charge = int(data['charge'])
        self.spin = int(data['spin'])
        self.props = data['props']
        self.rdmol = Chem.MolFromMolBlock(data['molblock'], sanitize=False, removeHs=False)

        return self


    def to_molblock(self, compressed: bool = False) -> str:
        """Returns MolBlock"""
        molblock = Chem.MolToMolBlock(self.rdmol)
        if compressed:
            molblock = compress_string(molblock)
        return molblock


    def to_xyz(self) -> str:
        """Returns XYZ formatted strings.

        Returns:
            str: XYZ formatted strings.
        """
        lines = [f'{self.natoms}', ' ']
        for e, (x, y, z) in zip(self.symbols(), self.positions()):
            lines.append(f'{e:5} {x:23.14f} {y:23.14f} {z:23.14f}')
        
        return '\n'.join(lines)


    def to_turbomole_coord(self, bohr: bool = False) -> str:
        """Returns TURBOMOLE coord file formatted strings.

        Turbomole coord file format:

            - It starts with the keyword `$coord`.
            - Each line after the $coord line specifies an atom, consisting of:
                - Three real numbers representing the Cartesian coordinates (x, y, z).
                - A string for the element name.
                - Optional: an "f" label at the end to indicate that the atom's coordinates are frozen during optimization.
            - Coordinates can be given in Bohr (default), Ångström (`$coord angs`), or fractional coordinates (`$coord frac`).
            - Optional data groups like periodicity (`$periodic`), lattice parameters (`$lattice`), and cell parameters (`$cell`) can also be included. 
            - Regarding precision:
                The precision of the coordinates is crucial for accurate calculations, especially geometry optimizations.
                Tools like the TURBOMOLEOptimizer might check for differences in atomic positions with a tolerance of 1e-13. 

        Args:
            bohr (bool): whether to use Bohr units of the coordinates. Defaults to False.
                Otherwise, Angstrom units will be used.
        
        Returns:
            str: TURBOMOLE coord formatted file.
        """
        if bohr:
            lines = ["$coord"]
        else:
            lines = ["$coord angs"]

        for (x, y, z), e in zip(self.positions(), self.symbols()):
            lines.append(f"{x:20.15f} {y:20.15f} {z:20.15f} {e}")
        
        lines.append("$end")

        return '\n'.join(lines)
        

    def to_sdf(self, props:bool=True) -> str:
        """Returns the SDF-formatted strings.

        Args:
            props (bool, optional): include `props as SDF properties. Defaults to True.

        Returns:
            str: strings in the SDF format.
        """
        in_memory = io.StringIO()
        with Chem.SDWriter(in_memory) as f:
            rdmol = Chem.Mol(self.rdmol)
            rdmol.SetProp('_Name', self.name)
            if props:
                for k,v in self.props.items():
                    rdmol.SetProp(k, str(v))
            f.write(rdmol)
        return in_memory.getvalue()


    def to_smi(self) -> str:
        """Returns SMILES containing explicit hydrogens

        Returns:
            str: SMILES
        """
        return Chem.MolToSmiles(self.rdmol)
    

    def to_png(self,
               width: int = 300, 
               height: int = 300,
               legend: str = '', 
               atom_index: bool = False, 
               highlight_atoms: list[int] | None = None,
               highlight_bonds: list[int] | None = None,
               redraw: bool = False,
               coordgen: bool = False,
               trim: bool = True) -> Image.Image:
        """Draw 2D molecule in PNG format.

        Args:
            width (int, optional): width. Defaults to 300.
            height (int, optional): height. Defaults to 300.
            legend (str, optional): legend. Defaults to ''.
            atom_index (bool, optional): whether to show atom index. Defaults to False.
            highlight_atoms (list[int] | None, optional): atom(s) to highlight. Defaults to None.
            highlight_bonds (list[int] | None, optional): bond(s) to highlight. Defaults to None.
            redraw (bool, optional): whether to redraw. Defaults to False.
            coordgen (bool, optional): whether to use coordgen. Defaults to False.
            trim (bool, optional): whether to trim white margins. Default to True.
        
        Returns:
            Image.Image: output PIL Image object.
        """
        
        return render_png(self.rdmol,
                          width = width, 
                          height = height, 
                          legend = legend, 
                          atom_index = atom_index, 
                          highlight_atoms = highlight_atoms,
                          highlight_bonds = highlight_bonds,
                          redraw = redraw, 
                          coordgen = coordgen, 
                          trim = trim)
    
    
    def to_svg(self,
               width: int = 300,
               height: int = 300,
               legend: str = '', 
               atom_index: bool = False, 
               highlight_atoms: list[int] | None = None,
               highlight_bonds: list[int] | None = None,
               redraw: bool = False,
               coordgen: bool = False,
               optimize: bool = True) -> str:
        """Draw 2D molecule in SVG format.

        Examples:
            For Jupyternotebook, wrap the output with SVG:

            >>> from IPython.display import SVG
            >>> SVG(libr[0].to_svg())

        Args:
            width (int, optional): width. Defaults to 300.
            height (int, optional): height. Defaults to 300.
            legend (str, optional): legend. Defaults to ''.
            atom_index (bool, optional): whether to show atom index. Defaults to False.
            highlight_atoms (list[int] | None, optional): atom(s) to highlight. Defaults to None.
            highlight_bonds (list[int] | None, optional): bond(s) to highlight. Defaults to None.
            redraw (bool, optional): whether to redraw. Defaults to False.
            coordgen (bool, optional): whether to use coordgen. Defaults to False.
            optimize (bool, optional): whether to optimize SVG string. Defaults to True.

        Returns:
            str: SVG string
        """
        return render_svg(self.rdmol, 
                          width = width, 
                          height = height, 
                          legend = legend, 
                          atom_index = atom_index, 
                          highlight_atoms = highlight_atoms,
                          highlight_bonds = highlight_bonds,
                          redraw = redraw, 
                          coordgen = coordgen, 
                          optimize = optimize)
    

def prep_confs_batches(confs: list[Conf], batchsize_atoms: int = 16384) -> list[namedtuple]:
    """Prepare batches of Chem.Mol conformers for batch optimization.

    - Each batch has up to `batchsize_atoms` number of atoms.
    - Each batch (namedtuple) has `rdmols`, `size`, `num_atoms` attributes.
    - Note that this function is not necessary for torsion energies 
        the batch-optimization of torsion energies are performed using internally prepared batches 
        of a series of conformers with different dihedral angles.

    Args:
        batchsize_atoms (int): max number of atoms in one batch. Defaults to 16384(=16*1024).

    Returns:
        list: list of batches.
    """
    Batch = namedtuple('Batch', ['rdmols', 'size', 'num_atoms'])
    batches = []
    batch_confs = []
    batch_atoms = 0
    for conf in confs:
        if (batch_atoms + conf.natoms) > batchsize_atoms:
            batches.append(Batch(rdmols=batch_confs, size=len(batch_confs), num_atoms=batch_atoms))
            # start over a new batch
            batch_confs = [ conf.rdmol ]
            batch_atoms = conf.natoms
        else:
            batch_confs.append(conf.rdmol)
            batch_atoms += conf.natoms
    if batch_atoms > 0: # last remaining batch
        batches.append(Batch(rdmols=batch_confs, size=len(batch_confs), num_atoms=batch_atoms))

    return batches


def batch_optimize(confs: list[Conf], calculator: Callable, **kwargs) -> list[Conf]:
    """Batch optimize a list of Conf objects using a given ASE calculator.

    Args:
        confs (list[Conf]): list of Conf objects to be optimized.
        calculator (Callable): batch optimizer.

    Returns:
        list[Conf]: list of optimized Conf objects.
    """
    batchsize_atoms = kwargs.get('batchsize_atoms', 16384)
    batches = prep_confs_batches(confs, batchsize_atoms=batchsize_atoms)
    optimized_confs = []
    for batch in batches:
        optimized = calculator(batch.rdmols).run(**kwargs)
        for rdmol in optimized.mols:
            conf = Conf(rdmol)
            conf.props = {
                'E_tot_init(kcal/mol)': float(rdmol.GetProp('E_tot_init(kcal/mol)')),
                'E_tot(kcal/mol)': float(rdmol.GetProp('E_tot(kcal/mol)')),
                'Converged': True if rdmol.GetProp('Converged') == 'True' else False,
            }
            optimized_confs.append(conf)
    
    return optimized_confs


def batch_singlepoint(confs: list[Conf], calculator: Callable, **kwargs) -> list[Conf]:
    """Batch singlepoint calculation for a list of Conf objects using a given ASE calculator.

    Args:
        confs (list[Conf]): list of Conf objects to be calculated.
        calculator (Callable): batch calculator.

    Returns:
        list[Conf]: list of Conf objects with updated properties.
    """
    batchsize_atoms = kwargs.get('batchsize_atoms', 16384)
    batches = prep_confs_batches(confs, batchsize_atoms=batchsize_atoms)
    calculated_confs = []
    for batch in batches:
        calculated = calculator(batch.rdmols).run(**kwargs)
        for rdmol in calculated.mols:
            conf = Conf(rdmol)
            conf.props = {
                'E_tot(kcal/mol)': float(rdmol.GetProp('E_tot(kcal/mol)')),
            }
            calculated_confs.append(conf)
    
    return calculated_confs


from xml.etree.ElementTree import Element,SubElement,Comment,tostring,parse
from xml.dom import minidom

A= Element('PAINS')
A.append(Comment('Filter Family A p.S23'))

B= Element('PAINSb')
B.append(Comment('Filter Family B S23-S25'))

C= Element('PAINSc')
C.append(Comment('Filter Family C S25-S37'))

with open('PAINS-more-than-150-hits.xml','rt') as f :
	treeA= parse(f)
with open('PAINS-less-than-150-hits.xml','rt') as g :
	treeB= parse(g)
with open('PAINS-less-than-015-hits.xml','rt') as h :
	treeC= parse(h)

count= 0
for node in treeA.iter() :
	if node.tag == 'SMART' :
		smarts= node.text.strip()
	if node.tag == 'name' :
		name= node.text.strip()
		count += 1
		entry= SubElement (A, 'group')
		entry.set('name', '('+str(count)+') '+name)
		node= SubElement (entry, "SMARTS")
		node.text= smarts

count= 0
for node in treeB.iter() :
	if node.tag == 'SMART' :
		smarts= node.text.strip()
	if node.tag == 'name' :
		name= node.text.strip()
		count += 1
		entry= SubElement (B, 'group')
		entry.set('name', '('+str(count)+') '+name)
		node= SubElement (entry, "SMARTS")
		node.text= smarts

count= 0
for node in treeC.iter() :
	if node.tag == 'SMART' :
		smarts= node.text.strip()
	if node.tag == 'name' :
		name= node.text.strip()
		count += 1
		entry= SubElement (C, 'group')
		entry.set('name', '('+str(count)+') '+name)
		node= SubElement (entry, "SMARTS")
		node.text= smarts
	

g=open('Baell2010A.xml','wt')
coarse= tostring(A,'utf-8')
g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )
g.close()

g=open('Baell2010B.xml','wt')
coarse= tostring(B,'utf-8')
g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )
g.close()

g=open('Baell2010C.xml','wt')
coarse= tostring(C,'utf-8')
g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )
g.close()



import numpy as np
import math
import itertools
import logging

from types import SimpleNamespace
from pathlib import Path
from rdworks import Conf, Mol
from rdworks.xtb.wrapper import GFN2xTB

from rdkit import Chem
from rdkit.Chem import (
    AllChem, RemoveHs, CanonSmiles, MolFromSmarts,
    GetFormalCharge, RWMol, AddHs, SanitizeMol, 
    MolToSmiles, MolFromSmiles,
    )

logger = logging.getLogger(__name__)


kT = 0.001987 * 298.0 # (kcal/mol K), standard condition
C = math.log(10) * kT



class Microstates():

    def __init__(self, origin: Mol, calculator: str = 'xTB'):
        self.origin = origin
        self.calculator = calculator
        self.basic_sites = []
        self.acidic_sites = []
        self.states = []
        self.mols = []
        self.reference = None
    

    def enumerate(self) -> None:
        # Qu pKake results must be stored at .confs
        for conf in self.origin:
            pka = conf.props.get('pka', None)
            if pka is None:
                # no protonation/deprotonation sites
                continue
            if isinstance(pka, str) and pka.startswith('tensor'):
                # ex. 'tensor(9.5784)'
                pka = float(pka.replace('tensor(','').replace(')',''))
            if conf.props.get('pka_type') == 'basic':
                self.basic_sites.append(conf.props.get('idx'))
            elif conf.props.get('pka_type') == 'acidic':
                self.acidic_sites.append(conf.props.get('idx'))

        # enumerate protonation/deprotonation sites to generate microstates

        np = len(self.basic_sites)
        nd = len(self.acidic_sites)
        P = [c for n in range(np+1) for c in itertools.combinations(self.basic_sites, n)]
        D = [c for n in range(nd+1) for c in itertools.combinations(self.acidic_sites, n)]
        
        PD = list(itertools.product(P, D))
        
        for (p, d) in PD:
            conf = self.origin.confs[0].copy()
            conf = conf.protonate(p).deprotonate(d).optimize(calculator=self.calculator)
            charge = len(p) - len(d)
            self.states.append(SimpleNamespace(
                charge=charge, 
                protonation_sites=p, 
                deprotonation_sites=d,
                conf=conf,
                smiles=Mol(conf).smiles,
                delta_m=None,
                PE=None))
            
        # sort microstates by ascending charges
        self.states = sorted(self.states, key=lambda x: x.charge)


    @staticmethod
    def Boltzmann_weighted_average(potential_energies: list) -> float:
        """Calculate Boltzmann weighted average potential energy at pH 0.

        Args:
            potential_energies (list): a list of potential energies.

        Returns:
            float: Boltzmann weighted average potential energy.
        """
        pe_array = np.array(potential_energies)
        pe = pe_array - min(potential_energies)
        Boltzmann_factors = np.exp(-pe/kT)
        Z = np.sum(Boltzmann_factors)
        p = Boltzmann_factors/Z

        return float(np.dot(p, pe_array))


    def populate(self) -> None:
        for microstate in self.states:
            mol = Mol(microstate.conf).make_confs(n=4).optimize_confs()
            # mol = mol.drop_confs(similar=True, similar_rmsd=0.3, verbose=True)
            # mol = mol.optimize_confs(calculator=calculator)
            # mol = mol.drop_confs(k=10, window=15.0, verbose=True)
            PE = []
            for conf in mol.confs:
                conf = conf.optimize(calculator=self.calculator, verbose=True)
                # GFN2xTB requires 3D coordinates
                # xtb = GFN2xTB(conf.rdmol).singlepoint(water='cpcmx', verbose=True)
                PE.append(conf.potential_energy(calculator=self.calculator))
                # xtb = GFN2xTB(conf.rdmol).singlepoint(verbose=True)
                # SimpleNamespace(
                #             PE = datadict['total energy'] * hartree2kcalpermol,
                #             Gsolv = Gsolv,
                #             charges = datadict['partial charges'],
                #             wbo = Wiberg_bond_orders,
                #             )
            microstate.PE = self.Boltzmann_weighted_average(PE)
            logger.info(f"PE= {PE}")
            logger.info(f"Boltzmann weighted= {microstate.PE}")            
            self.mols.append(mol)


    def get_populations(self, pH: float) -> list[tuple]:
        # set the lowest dG as the reference
        self.reference = self.states[np.argmin([microstate.PE for microstate in self.states])]
        for microstate in self.states:
            microstate.delta_m = microstate.charge - self.reference.charge
        dG = []
        for microstate in self.states:
            dG.append((microstate.PE - self.reference.PE) + microstate.delta_m * C * pH)
        dG = np.array(dG)

        logger.info(f"dG= {dG}")
        Boltzmann_factors = np.exp(-dG/kT)
        Z = np.sum(Boltzmann_factors)
        p = Boltzmann_factors/Z
        idx_p = sorted(list(enumerate(p)), key=lambda x: x[1], reverse=True)
        # [(0, p0), (1, p1), ...]

        return idx_p


    def get_ensemble(self) -> list[Mol]:
        return self.mols


    def get_mol(self, idx: int) -> Mol:
        return self.mols[idx]
    

    def count(self) -> int:
        return len(self.states)
    



from __future__ import print_function
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.dom import minidom
from rdkit import Chem
import sys

with open("hann1999filter.txt","r") as f, \
     open("Hann1999.xml","w") as g:
  root= Element('Hann1999')
  for line in f :
    if line.startswith("#") : continue 
    c= line.strip().split()
    if len(c) < 2 : continue
    try:
      smartstxt= c[-1]
      name= " ".join(c[:-1])
    except:
      print(line.rstrip())
      sys.exit(0)
    try:
      rdkq= Chem.MolFromSmarts(smartstxt)
    except:
      print("NOT VALID:", line.rstrip())
      continue
    entry= SubElement (root, 'group')
    entry.set('name', name)
    node= SubElement (entry, 'SMARTS')
    node.text= smartstxt
  # write to XML
  coarse= tostring(root,'utf-8')
  g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )

with open("hann1999acid.txt","r") as f, \
     open("Hann1999Acid.xml","w") as g:
  root= Element('Hann1999')
  for line in f :
    if line.startswith("#") : continue 
    c= line.strip().split()
    if len(c) < 2 : continue
    try:
      smartstxt= c[-1]
      name= " ".join(c[:-1])
    except:
      print(line.rstrip())
      sys.exit(0)
    try:
      rdkq= Chem.MolFromSmarts(smartstxt)
    except:
      print("NOT VALID:", line.rstrip())
      continue
    entry= SubElement (root, 'group')
    entry.set('name', name)
    node= SubElement (entry, 'SMARTS')
    node.text= smartstxt
  # write to XML
  coarse= tostring(root,'utf-8')
  g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )

with open("hann1999acid.txt","r") as f, \
     open("Hann1999Acid.xml","w") as g:
  root= Element('Hann1999')
  for line in f :
    if line.startswith("#") : continue 
    c= line.strip().split()
    if len(c) < 2 : continue
    try:
      smartstxt= c[-1]
      name= " ".join(c[:-1])
    except:
      print(line.rstrip())
      sys.exit(0)
    try:
      rdkq= Chem.MolFromSmarts(smartstxt)
    except:
      print("NOT VALID:", line.rstrip())
      continue
    entry= SubElement (root, 'group')
    entry.set('name', name)
    node= SubElement (entry, 'SMARTS')
    node.text= smartstxt
  # write to XML
  coarse= tostring(root,'utf-8')
  g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )



import os
import pathlib
import copy
import operator
import numpy as np
import math

from collections import defaultdict
from typing import List, Tuple, Union, Iterator

from rdkit import Chem, Geometry
from rdkit.Chem import Draw, AllChem, rdMMPA, rdDepictor

from rdworks.mol import Mol, rd_descriptor, rd_descriptor_f
from rdworks.mollibr import MolLibr
from rdworks.display import render_svg


def get_attachment_orientation(mol: Chem.Mol, dummy_atom_idx: int | None = None):
    """
    Get the orientation of attachment points (dummy atoms) in a molecule.
    
    Args:
        mol: RDKit molecule object
        dummy_atom_idx: Index of specific dummy atom (if None, finds all dummy atoms)
    
    Returns:
        Dictionary with dummy atom indices as keys and orientation info as values
    """
    # Ensure molecule has 2D coordinates
    if not mol.GetNumConformers():
        rdDepictor.Compute2DCoords(mol)
    
    conf = mol.GetConformer()
    orientations = {}
    
    # Find dummy atoms (atomic number 0)
    dummy_atoms = []
    if dummy_atom_idx is not None:
        dummy_atoms = [dummy_atom_idx]
    else:
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 0:  # Dummy atom
                dummy_atoms.append(atom.GetIdx())
    
    for dummy_idx in dummy_atoms:
        dummy_atom = mol.GetAtomWithIdx(dummy_idx)
        dummy_pos = np.array(conf.GetAtomPosition(dummy_idx))[:2]  # Get x, y coordinates
        
        # Get the neighboring atom (should be exactly one for attachment point)
        neighbors = list(dummy_atom.GetNeighbors())
        if len(neighbors) != 1:
            print(f"Warning: Dummy atom {dummy_idx} has {len(neighbors)} neighbors")
            continue
            
        neighbor_idx = neighbors[0].GetIdx()
        neighbor_pos = np.array(conf.GetAtomPosition(neighbor_idx))[:2]
        
        # Calculate direction vector from neighbor to dummy atom
        direction_vector = dummy_pos - neighbor_pos
        
        # Normalize the vector
        if np.linalg.norm(direction_vector) > 0:
            direction_vector = direction_vector / np.linalg.norm(direction_vector)
        
        # Calculate angle in degrees (0° = right, 90° = up, 180° = left, 270° = down)
        angle_rad = math.atan2(direction_vector[1], direction_vector[0])
        angle_deg = math.degrees(angle_rad)
        
        # Normalize angle to 0-360 range
        if angle_deg < 0:
            angle_deg += 360
            
        # Determine cardinal direction
        if angle_deg <= 45 or angle_deg > 315:
            cardinal = "right"
        elif angle_deg <= 135:
            cardinal = "up"
        elif angle_deg <= 225:
            cardinal = "left"
        else:
            cardinal = "down"
        
        orientations[dummy_idx] = {
            'angle_deg': angle_deg,
            'direction_vector': direction_vector,
            'cardinal': cardinal,
            'neighbor_idx': neighbor_idx,
            'dummy_pos': dummy_pos,
            'neighbor_pos': neighbor_pos
        }
    
    return orientations



def orient_rgroup_to_attachment(rgroup_mol: Chem.Mol, scaffold_orientations: dict) -> Chem.Mol:
    """
    Orient an R-group molecule to match the attachment point orientation of a scaffold.
    
    Args:
        scaffold_mol: Scaffold molecule with dummy atom
        rgroup_mol: R-group molecule with dummy atom
        scaffold_dummy_idx: Index of dummy atom in scaffold
        rgroup_dummy_idx: Index of dummy atom in R-group
    
    Returns:
        Oriented R-group molecule
    """
    # Get orientations
    scaffold_dummy_atom_idx = list(scaffold_orientations.keys())
    if len(scaffold_dummy_atom_idx) == 0:
        print("Error: No dummy atoms found in scaffold molecule")
        return rgroup_mol
    
    rgroup_orientations = get_attachment_orientation(rgroup_mol)

    rgroup_dummy_atom_idx = list(rgroup_orientations.keys())
    if len(rgroup_dummy_atom_idx) == 0:
        print("Error: No dummy atoms found in R-group molecule")
        return rgroup_mol
    
    scaffold_info = scaffold_orientations[scaffold_dummy_atom_idx[0]]
    rgroup_info = rgroup_orientations[rgroup_dummy_atom_idx[0]]
    # Calculate rotation needed
    # We want the R-group attachment to point in the opposite direction of scaffold attachment
    target_angle = scaffold_info['angle_deg'] + 180  # Opposite direction
    current_angle = rgroup_info['angle_deg']
    rotation_angle = target_angle - current_angle
    
    # copy of input molecule
    oriented_mol = Chem.Mol(rgroup_mol)
    
    # Apply rotation to R-group coordinates
    rdDepictor.Compute2DCoords(oriented_mol)
    conf = oriented_mol.GetConformer()
    
    # Get center of rotation (R-group dummy atom position)
    center = rgroup_info['dummy_pos']
    
    # Rotate all atoms
    for i in range(oriented_mol.GetNumAtoms()):
        pos = np.array(conf.GetAtomPosition(i))[:2]
        
        # Translate to origin
        pos_centered = pos - center
        
        # Apply rotation
        angle_rad = math.radians(rotation_angle)
        cos_a, sin_a = math.cos(angle_rad), math.sin(angle_rad)
        rotation_matrix = np.array([[cos_a, -sin_a], [sin_a, cos_a]])
        pos_rotated = np.dot(rotation_matrix, pos_centered)
        
        # Translate back
        pos_final = pos_rotated + center
        
        # Update coordinates
        conf.SetAtomPosition(i, (pos_final[0], pos_final[1], 0.0))
    
    return oriented_mol


def calculate_bond_length(mol: Chem.Mol, atom1_idx: int, atom2_idx: int) -> float:
    """
    Calculate the 2D distance between two bonded atoms.
    
    Args:
        mol: RDKit molecule
        atom1_idx: Index of first atom
        atom2_idx: Index of second atom
    
    Returns:
        Bond length in 2D coordinates
    """
    if not mol.GetNumConformers():
        rdDepictor.Compute2DCoords(mol)
    
    conf = mol.GetConformer()
    pos1 = np.array(conf.GetAtomPosition(atom1_idx))[:2]
    pos2 = np.array(conf.GetAtomPosition(atom2_idx))[:2]
    
    return np.linalg.norm(pos2 - pos1)



def get_average_bond_length(mol: Chem.Mol, bond_type: str = "C-C") -> float:
    """
    Calculate the average bond length for a specific bond type in a molecule.
    
    Args:
        mol: RDKit molecule
        bond_type: Type of bond to measure (e.g., "C-C", "C-N", "any")
    
    Returns:
        Average bond length
    """
    if not mol.GetNumConformers():
        rdDepictor.Compute2DCoords(mol)
    
    bond_lengths = []
    
    for bond in mol.GetBonds():
        atom1 = bond.GetBeginAtom()
        atom2 = bond.GetEndAtom()
        
        # Filter by bond type if specified
        if bond_type == "any":
            include_bond = True
        elif bond_type == "C-C":
            include_bond = (atom1.GetSymbol() == "C" and atom2.GetSymbol() == "C")
        elif bond_type == "C-N":
            include_bond = ((atom1.GetSymbol() == "C" and atom2.GetSymbol() == "N") or
                          (atom1.GetSymbol() == "N" and atom2.GetSymbol() == "C"))
        else:
            # Parse custom bond type like "C-O"
            symbols = bond_type.split("-")
            if len(symbols) == 2:
                include_bond = ((atom1.GetSymbol() == symbols[0] and atom2.GetSymbol() == symbols[1]) or
                              (atom1.GetSymbol() == symbols[1] and atom2.GetSymbol() == symbols[0]))
            else:
                include_bond = False
        
        if include_bond:
            length = calculate_bond_length(mol, atom1.GetIdx(), atom2.GetIdx())
            bond_lengths.append(length)
    
    return np.mean(bond_lengths) if bond_lengths else 0.0


def scale_molecule_coordinates(mol: Chem.Mol, target_bond_length: float, reference_bond_type: str = "any"):
    """
    Scale molecule coordinates to achieve a target bond length.
    
    Args:
        mol: RDKit molecule to scale
        target_bond_length: Desired bond length
        reference_bond_type: Bond type to use as reference for scaling
    
    Returns:
        New molecule with scaled coordinates
    """
    # copy of input molecule
    if not isinstance(mol, Chem.Mol):
        raise TypeError("Input must be an RDKit Mol object")
    
    scaled_mol = Chem.Mol(mol)
    
    # Ensure 2D coordinates exist
    if not scaled_mol.GetNumConformers():
        rdDepictor.Compute2DCoords(scaled_mol)
    
    # Calculate current average bond length
    current_avg_length = get_average_bond_length(scaled_mol, reference_bond_type)
    
    if current_avg_length == 0:
        # print(f"Warning: No {reference_bond_type} bonds found in molecule")
        return scaled_mol
    
    # Calculate scaling factor
    scale_factor = target_bond_length / current_avg_length
    
    # Apply scaling to all coordinates
    conf = scaled_mol.GetConformer()
    for i in range(scaled_mol.GetNumAtoms()):
        pos = conf.GetAtomPosition(i)
        new_pos = (pos.x * scale_factor, pos.y * scale_factor, pos.z)
        conf.SetAtomPosition(i, new_pos)
    
    return scaled_mol



def standardize_rgroup_scales(molecules: list[Chem.Mol], 
                              target_bond_length: float | None = None,
                              reference_bond_type: str = "C-C") -> list[Chem.Mol]:
    """
    Standardize the scale of multiple R-group molecules to have identical bond lengths.
    
    Args:
        molecules: List of RDKit molecules
        target_bond_length: Target bond length (if None, uses average across all molecules)
        reference_bond_type: Bond type to use as reference
    
    Returns:
        List of molecules with standardized scales
    """
    # Ensure all molecules have 2D coordinates
    for mol in molecules:
        if not mol.GetNumConformers():
            rdDepictor.Compute2DCoords(mol)
    
    # Calculate target bond length if not provided
    if target_bond_length is None:
        all_avg_lengths = []
        for mol in molecules:
            avg_length = get_average_bond_length(mol, reference_bond_type)
            if avg_length > 0:
                all_avg_lengths.append(avg_length)
        
        if all_avg_lengths:
            target_bond_length = np.mean(all_avg_lengths)
        else:
            # print(f"Warning: No {reference_bond_type} bonds found in any molecule")
            target_bond_length = 1.0  # Default value
    
    # Scale all molecules
    scaled_molecules = []
    for mol in molecules:
        scaled_mol = scale_molecule_coordinates(mol, target_bond_length, reference_bond_type)
        scaled_molecules.append(scaled_mol)
    
    return scaled_molecules


def create_uniform_grid_image(molecules: list[Chem.Mol],
                              legends: list[str] | None = None,
                              mols_per_row: int = 4,
                              sub_img_size: tuple[int, int] = (300, 300),
                              target_bond_length: float | None = None,
                              reference_bond_type: str = "C-C"):
    """
    Create a grid image of molecules with uniform scaling.
    
    Args:
        molecules: List of RDKit molecules
        legends: Optional list of molecule names/labels
        mols_per_row: Number of molecules per row
        sub_img_size: Size of each molecule image
        target_bond_length: Target bond length for scaling
        reference_bond_type: Bond type to use as reference
    
    Returns:
        PIL Image object
    """
    # Standardize scales
    scaled_molecules = standardize_rgroup_scales(molecules, 
                                                 target_bond_length, 
                                                 reference_bond_type)
    
    # Create grid image
    img = Draw.MolsToGridImage(
        scaled_molecules,
        molsPerRow = mols_per_row,
        subImgSize = sub_img_size,
        legends = legends
    )
    
    return img



class MatchedSeries:
    def __init__(self, 
                 mollibr:MolLibr,
                 sort_props:Union[List,str,None]=None,
                 core_min:int=5, core_max:int=30, size_min:int=3) -> None :
        """Initialize.

        Documented here: [MMS with rdkit](https://iwatobipen.wordpress.com/2016/02/01/create-matched-molecular-series-with-rdkit/),
        [Mishima-syk](https://github.com/Mishima-syk/py4chemoinformatics/blob/master/ch07_graph.asciidoc), 
        and [rdkit docs](http://rdkit.org/docs/source/rdkit.Chem.rdMMPA.html).            

        Examples:
            >>> import rdworks
            >>> libr = rdworks.read_smi('test.smi')
            >>> series = rdworks.MatchedSeries(libr)
        
        Args:
            mollibr (MolLibr): a library of molecules.
            sort_props (Union[List,str,None], optional): how to sort molecules within a series. Defaults to None.
            core_min (int, optional): min number of atoms for a core. Defaults to 5.
            core_max (int, optional): max number of atoms for a core. Defaults to 30.
            size_min (int, optional): min number of molecules for a series. Defaults to 3.

        Raises:
            TypeError: if `mollibr` is not rdworks.MolLibr object.
        """
        if isinstance(mollibr, MolLibr):
            self.mollibr = copy.deepcopy(mollibr) # a copy of MolLibr
        else:
            raise TypeError('MatchedSeries() expects rdworks.MolLibr object')
        if isinstance(sort_props, list):
            self.sort_props = sort_props
        elif isinstance(sort_props, str):
            self.sort_props = [ sort_props ]
        else:
            self.sort_props = [ 'HAC' ]
        self.core_min = core_min
        self.core_max = core_max
        self.size_min = size_min # minimum numer of R-groups in a series
        # for consistent drawing
        self.template_pattern = None
        self.template_coord2D = None
        self.series = self.libr_to_series() 
        
        
    def __str__(self) -> str:
        """Returns a string representation of object.

        Returns:
            str: string representation.
        """
        return f"<rdworks.MatchedSeries({self.count()})>"
    

    def __iter__(self) -> Iterator:
        """Yields an iterator of molecules.

        Yields:
            Iterator: iterator of molecules.
        """
        return iter(self.series)
    
    
    def __next__(self) -> Tuple:
        """Next series.

        Returns:
            Tuple: (scaffold_SMILES, [(r-group_SMILES, rdworks.Mol, *sort_props_values)
        """
        return next(self.series)
    

    def __getitem__(self, index:Union[int,slice]) -> Tuple:
        """Operator `[]`.

        Args:
            index (Union[int,slice]): index or indexes.

        Raises:
            ValueError: if series is empty or index is out of range.

        Returns:
            Tuple: (scaffold_SMILES, [(r-group_SMILES, rdworks.Mol, *sort_props_values)
        """
        if self.count() == 0:
            raise ValueError(f"MatchedSeries is empty")
        try:
            return self.series[index]
        except:
            raise ValueError(f"index should be 0..{self.count()-1}")
        

    def count(self) -> int:
        """Returns the count of series.

        Returns:
            int: count of series.
        """
        return len(self.series)
    

    def libr_to_series(self) -> List[Tuple]:
        """Returns a list of molecular series.

        Raises:
            RuntimeError: if a molecular cut cannot be defined.

        Returns:
            List[Tuple]: 
                [
                (scaffold_SMILES, [(r-group_SMILES, rdworks.Mol, *sort_props_values), ...,]),  
                ...,
                ]
        """
        series = defaultdict(list)
        for mol in self.mollibr:
            # make a single cut
            list_of_frag = rdMMPA.FragmentMol(mol.rdmol, maxCuts=1, resultsAsMols=False)
            # note: default parameters: maxCuts=3, maxCutBonds=20, resultsAsMols=True
            for _, cut in list_of_frag:
                try:
                    frag_smiles_1, frag_smiles_2 = cut.split('.')
                except:
                    raise RuntimeError(f'{mol.name} fragment_tuple= {cut}')
                n1 = Chem.MolFromSmiles(frag_smiles_1).GetNumHeavyAtoms()
                n2 = Chem.MolFromSmiles(frag_smiles_2).GetNumHeavyAtoms()
                # split scaffold core and rgroup symmetrically
                if n1 >= self.core_min and n1 <= self.core_max and n1 > n2: 
                    # frag_1 is the scaffold and frag_2 is the rgroup
                    series[frag_smiles_1].append((frag_smiles_2, mol))
                if n2 >= self.core_min and n2 <= self.core_max and n2 > n1: 
                    # frag_2 is the scaffold and frag_1 is the rgroup
                    series[frag_smiles_2].append((frag_smiles_1, mol))    
        # convert dict to list and remove size < self.size_min
        series = [(k,v) for k,v in series.items() if len(v) >= self.size_min]
        # sort by size (from the largest to the smallest)
        series = sorted(series, key=lambda x: len(x[1]), reverse=True)
        # sort by self.sort_props
        series_r_group_sorted = []
        for (scaffold_smi, r_group_) in series:
            r_group = []
            for (r_smi, mol) in r_group_:
                values = []
                for p in self.sort_props:
                    try:
                        v = mol.props[p]
                    except:
                        if p in rd_descriptor_f:
                            v = rd_descriptor_f[p](mol.rdmol) # calc. on the fly
                            mol.props.update({p:v})
                        else:
                            v = None
                    values.append(v)
                r_group.append((r_smi, mol, *values)) # unpack values i.e. a=[2,3] b=(1,*a) == (1,2,3)
            r_group = sorted(r_group, key=operator.itemgetter(slice(2, 2+len(self.sort_props))))
            series_r_group_sorted.append((scaffold_smi, r_group))
        return series_r_group_sorted
    

    def template(self, SMARTS:str, rdmol:Chem.Mol) -> None:
        """Sets drawing layout template.

        Args:
            SMARTS (str): SMARTS for template pattern.
            rdmol (Chem.Mol): template molecule.
        """
        
        self.template_pattern = Chem.MolFromSmarts(SMARTS)
        matched = rdmol.GetSubstructMatch(self.template_pattern)
        coords = [rdmol.GetConformer().GetAtomPosition(x) for x in matched]
        self.template_coords2D = [Geometry.Point2D(pt.x, pt.y) for pt in coords]


    def depict(self, smiles:str) -> Chem.Mol:
        """Draws a molecule according to self.template in a consistent way.

        Args:
            smiles (str): input molecule.

        Returns:
            Chem.Mol: 2D coordinated Chem.Mol for depiction.
        """
        rdmol_2d = Chem.MolFromSmiles(smiles)
        try:
            matched = rdmol_2d.GetSubstructMatch(self.template_pattern)
            coordDict = {}
            for i, coord in enumerate(self.template_coords2D):
                coordDict[matched[i]] = coord
            AllChem.Compute2DCoords(rdmol_2d, coordMap=coordDict)
        except:
            pass
        return rdmol_2d


    def report(self,
               workdir: os.PathLike = pathlib.Path("."),
               prefix: str = "mmseries",
               num_columns: int = 5,
               rgroup_width: int = 200,
               rgroup_height: int = 200,
               rgroup_max: int | None = None,
               scaffold_width: int = 300,
               scaffold_height: int = 300,
               ) -> None:
        """Writes individual series and an overview of series as HTML files.

        Args:
            workdir (os.PathLike, optional): working directory. Defaults to pathlib.Path(".").
            prefix (str, optional): prefix of output files. Defaults to "mmseries".
            num_columns (int, optional): number of molecules per row. Defaults to 5.
            rgroup_width (int, optional): width. Defaults to 200.
            rgroup_height (int, optional): height. Defaults to 200.
            rgroup_max (int, optional): max number of R-group molecules. Defaults to None.
            scaffold_width (int, optional): width. Defaults to 300.
            scaffold_height (int, optional): height. Defaults to 300.
        """
        scaffold_mols = []
        scaffold_legends = []
        scaffold_svgs = []
        
        for idx, (scaffold_smiles, list_tuples_r_groups) in enumerate(self.series, start=1):
            num = len(list_tuples_r_groups)
            scaffold_mols.append(Chem.MolFromSmiles(scaffold_smiles))
            scaffold_legends.append(f'Series #{idx} (n={num})')
            r_group_mols = []
            r_group_legends = []
            for (r_group_smiles, m, *values) in list_tuples_r_groups:
                # (r-group_SMILES, rdworks.Mol, *sort_props_values)
                values = list(map(str, values))
                r_group_mols.append(Chem.MolFromSmiles(r_group_smiles))
                r_group_legends.append(f'{m.name}\n{",".join(values)}')

            # individual series
            scaffold_svgs.append(render_svg(scaffold_mols[-1], 
                                        width = scaffold_width, 
                                        height = scaffold_height, 
                                        legend = scaffold_legends[-1]))
            
            # new HTML file
            HTML = '<html>'
            HTML += '<head>'
            HTML += '<meta charset="UTF-8">'
            HTML += '<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">'
            HTML += '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@4.0.0/dist/css/bootstrap.min.css" integrity="sha384-Gn5384xqQ1aoWXA+058RXPxPg6fy4IWvTNh0E263XmFcJlSAwiGgFAW/dAiS6JXm" crossorigin="anonymous">'
            HTML += '<title>Matched Series Report</title>'
            HTML += '</head>'
            HTML += '<body>'
            HTML += f'<div>{scaffold_svgs[-1]}</div>'
            
            scaffold_orientations = get_attachment_orientation(scaffold_mols[-1])
            # orient r-group to scaffold attachment
            r_group_mols = [orient_rgroup_to_attachment(rg, scaffold_orientations) for rg in r_group_mols]
            # scale r-group coordinates
            r_group_mols = standardize_rgroup_scales(r_group_mols, 
                                                     target_bond_length=1.2,
                                                     reference_bond_type="C-C")
            HTML += '<table class="table table-bordered">'
            HTML += '<tbody>'
            HTML += '<tr>'
            truncated = False
            for i, (mol, legend) in enumerate(zip(r_group_mols, r_group_legends), start=1):
                if rgroup_max is not None and i > rgroup_max:
                    truncated = True
                    break
                rgroup_svg = render_svg(mol,
                                        width = rgroup_width, 
                                        height = rgroup_height, 
                                        legend = legend)
                HTML += f'<td>{rgroup_svg}</td>'
                if i % num_columns == 0:
                    HTML += '</tr><tr>'
            if i % num_columns != 0:
                HTML += '</tr>'
            HTML += '</tbody></table>'
            if truncated:
                HTML += f'<p>Note: R-groups are truncated to {rgroup_max} molecules</p>'
            HTML += '<script src="https://code.jquery.com/jquery-3.2.1.slim.min.js" integrity="sha384-KJ3o2DKtIkvYIK3UENzmM7KCkRr/rE9/Qpg6aAZGJwFDMVNA/GpGFF93hXpG5KkN" crossorigin="anonymous"></script>'
            HTML += '<script src="https://cdn.jsdelivr.net/npm/popper.js@1.12.9/dist/umd/popper.min.js" integrity="sha384-ApNbgh9B+Y1QKtv3Rn7W3mgPxhU9K/ScQsAP7hUibX39j7fakFPskvXusvfa0b4Q" crossorigin="anonymous"></script>'
            HTML += '<script src="https://cdn.jsdelivr.net/npm/bootstrap@4.0.0/dist/js/bootstrap.min.js" integrity="sha384-JZR6Spejh4U02d8jOt6vLEHfe/JQGiRRSQQxSfFWpi1MquVdAyjUar5+76PVCmYl" crossorigin="anonymous"></script>'
            HTML += '</body>'
            HTML += '</html>'

            # individual series
            with open(workdir / f"{prefix}-{idx:03d}-count-{num:03d}.html", "w") as html:    
                html.write(HTML)

        # overview of scaffolds
        HTML = '<html>'
        HTML += '<head>'
        HTML += '<meta charset="UTF-8">'
        HTML += '<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">'
        HTML += '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@4.0.0/dist/css/bootstrap.min.css" integrity="sha384-Gn5384xqQ1aoWXA+058RXPxPg6fy4IWvTNh0E263XmFcJlSAwiGgFAW/dAiS6JXm" crossorigin="anonymous">'
        HTML += '<title>Matched Series Overview Report</title>'
        HTML += '</head>'
        HTML += '<body>'
        HTML += '<table class="table table-bordered">'
        HTML += '<tbody>'
        HTML += '<tr>'
        for i, svg in enumerate(scaffold_svgs, start=1):
            HTML += f'<td>{svg}</td>'
            if i % num_columns == 0:
                HTML += '</tr><tr>'
        if i % num_columns != 0:
            HTML += '</tr>'
        HTML += '</tbody></table>'
        HTML += '<script src="https://code.jquery.com/jquery-3.2.1.slim.min.js" integrity="sha384-KJ3o2DKtIkvYIK3UENzmM7KCkRr/rE9/Qpg6aAZGJwFDMVNA/GpGFF93hXpG5KkN" crossorigin="anonymous"></script>'
        HTML += '<script src="https://cdn.jsdelivr.net/npm/popper.js@1.12.9/dist/umd/popper.min.js" integrity="sha384-ApNbgh9B+Y1QKtv3Rn7W3mgPxhU9K/ScQsAP7hUibX39j7fakFPskvXusvfa0b4Q" crossorigin="anonymous"></script>'
        HTML += '<script src="https://cdn.jsdelivr.net/npm/bootstrap@4.0.0/dist/js/bootstrap.min.js" integrity="sha384-JZR6Spejh4U02d8jOt6vLEHfe/JQGiRRSQQxSfFWpi1MquVdAyjUar5+76PVCmYl" crossorigin="anonymous"></script>'
        HTML += '</body>'
        HTML += '</html>'

        with open(workdir / f"{prefix}-overview.html", "w") as html:
            html.write(HTML)


import importlib.resources
from types import SimpleNamespace
import pandas as pd

from rdkit import Chem

# adapted from https://github.com/dptech-corp/Uni-pKa/enumerator

class IonizedStates:
    """Knowledge-based enumeration of (de)protonated states"""

    smarts_path = importlib.resources.files('rdworks.predefined.ionized')
    ionization_patterns = pd.read_csv(smarts_path / 'simple_smarts_pattern.csv')
    
    # Unreasonable chemical structures
    unreasonable_patterns = [
        Chem.MolFromSmarts(s) for s in [
            "[#6X5]",
            "[#7X5]",
            "[#8X4]",
            "[*r]=[*r]=[*r]",
            "[#1]-[*+1]~[*-1]",
            "[#1]-[*+1]=,:[*]-,:[*-1]",
            "[#1]-[*+1]-,:[*]=,:[*-1]",
            "[*+2]",
            "[*-2]",
            "[#1]-[#8+1].[#8-1,#7-1,#6-1]",
            "[#1]-[#7+1,#8+1].[#7-1,#6-1]",
            "[#1]-[#8+1].[#8-1,#6-1]",
            "[#1]-[#7+1].[#8-1]-[C](-[C,#1])(-[C,#1])",
            # "[#6;!$([#6]-,:[*]=,:[*]);!$([#6]-,:[#7,#8,#16])]=[C](-[O,N,S]-[#1])",
            # "[#6]-,=[C](-[O,N,S])(-[O,N,S]-[#1])",
            "[OX1]=[C]-[OH2+1]",
            "[NX1,NX2H1,NX3H2]=[C]-[O]-[H]",
            "[#6-1]=[*]-[*]",
            "[cX2-1]",
            "[N+1](=O)-[O]-[H]",
        ]]


    def __init__(self, smiles: str, charge_min: int = -2, charge_max: int = 2):
        self.smiles = Chem.CanonSmiles(smiles)
        self.charge_max = charge_max
        self.charge_min = charge_min

        self.rdmol = Chem.MolFromSmiles(self.smiles)
        self.rdmol_H = Chem.AddHs(self.rdmol)
        self.charge = Chem.GetFormalCharge(self.rdmol_H)
        
        # initial states
        self.states = {self.smiles : (self.rdmol_H, self.charge)}
        
        # initial ionization sites
        self.sites = {self.smiles: self.set_ionization_sites(self.smiles)}

        # pKa pairs: 
        # HA(acid) + H2O == A-(base) + H3O+ or HA+(acid) + H2O == A(base) + H3O+
        self.pairs = []

        # iteratively build an ensemble of ionized states
        self.ensemble()


    
    @staticmethod
    def set_ionization_sites(smiles: str) -> tuple:
        subject = Chem.MolFromSmiles(smiles)
        subject = Chem.AddHs(subject)
        charge = Chem.GetFormalCharge(subject)        
        indices = [] # atom indices of protonation/deprotonation site(s)
        for i, name, smarts, smarts_index, acid_or_base in IonizedStates.ionization_patterns.itertuples():
            pattern = Chem.MolFromSmarts(smarts)
            matches = subject.GetSubstructMatches(pattern)
            # returns a list of tuples, where each tuple contains the indices 
            # of the atoms in the molecule that match the substructure query
            # ex. ((1,), (2,), (3,))
            if len(matches) > 0:
                smarts_index = int(smarts_index)
                indices += [(match[smarts_index], acid_or_base) for match in matches]
        
        return (list(set(indices)), subject, charge)
    

    @staticmethod
    def clean_smiles(rdmol: Chem.Mol) -> str:
        Chem.SanitizeMol(rdmol)
        rdmol = Chem.MolFromSmiles(Chem.MolToSmiles(rdmol))
        rdmol_H = Chem.AddHs(rdmol)
        rdmol = Chem.RemoveHs(rdmol_H)
        return Chem.CanonSmiles(Chem.MolToSmiles(rdmol))


    @staticmethod
    def reasonable(romol: Chem.Mol) -> bool:
        return all([len(romol.GetSubstructMatches(p)) == 0 for p in IonizedStates.unreasonable_patterns])
    

    def ionize(self, smiles: str | None = None) -> int:
        num_added_states = 0
        
        if smiles is None:
            smiles = self.smiles

        if smiles not in self.sites:
            self.sites[smiles] = self.set_ionization_sites(smiles)
        
        (indices, subject, charge) = self.sites[smiles]
        
        if (charge >= self.charge_max) or (charge <= self.charge_min):
            # formal charge will be increased or decreased by protonation/deprotonation
            # so, if the charge of current state is already max or min
            # there is nothing to do
            return num_added_states
            
        for (i, acid_or_base) in indices:
            edmol = Chem.RWMol(subject) # edmol preserves Hs
            if acid_or_base == 'A': # de-protonate
                A = edmol.GetAtomWithIdx(i)
                if A.GetAtomicNum() == 1:
                    X = A.GetNeighbors()[0] # there must be only one neighbor
                    charge = X.GetFormalCharge() - 1
                    X.SetFormalCharge(charge)
                    edmol.RemoveAtom(i)
                else:
                    bonded_H_indices = [ H.GetIdx() for H in A.GetNeighbors() if H.GetAtomicNum() == 1 ]
                    nH = len(bonded_H_indices)
                    assert nH > 0, f"Cannot deprotonate an atom (idx={i}; no H)"
                    charge = A.GetFormalCharge() - 1
                    A.SetFormalCharge(charge)
                    edmol.RemoveAtom(bonded_H_indices[0])
            
            elif acid_or_base == 'B': # protonate
                # note that protonation at tertiary nitrogen may results in stereoisomers
                # current implementation ignores the stereochemistry
                # use rdworks.complete_stereoisomers() function to complete the stereoisomers
                B = edmol.GetAtomWithIdx(i)
                assert B.GetAtomicNum() > 1, f"Cannot protonate an atom (idx={i}; {B.GetAtomicNum()})"
                charge = B.GetFormalCharge() + 1
                B.SetFormalCharge(charge)
                nH = B.GetNumExplicitHs()
                B.SetNumExplicitHs(nH+1)
                edmol = Chem.AddHs(edmol)
            
            # clean up and save SMILES
            ionized_smiles = IonizedStates.clean_smiles(edmol)
            ionized_mol = Chem.MolFromSmiles(ionized_smiles)
            ionized_mol = Chem.AddHs(ionized_mol)
            ionized_charge = Chem.GetFormalCharge(ionized_mol)
            if self.reasonable(ionized_mol):
                if ionized_smiles in self.states:
                    continue
                self.states[ionized_smiles] = (ionized_mol, ionized_charge)
                num_added_states += 1
            
                # store acid-base pair information for pKa
                if acid_or_base == 'A':
                    self.pairs.append((i, smiles, ionized_smiles))
                elif acid_or_base == 'B':
                    self.pairs.append((i, ionized_smiles, smiles))

        return num_added_states
    
    
    def ensemble(self) -> None:
        # populate initial states
        self.ionize()

        # propagate
        num_added_states = None       
        while num_added_states is None or num_added_states > 0:
            states = self.states.copy() # dictionary
            # self.ionize(smiles) below will change self.states
            # so we cannot iterate self.states. Instead we will
            # iterate over a copy of the self.states
            for smiles in states:
                num_added_states = self.ionize(smiles) 
                
    
    def count(self) -> int:
        return len(self.states)


    def get_sites(self) -> dict:
        return self.sites
    
    
    def get_smiles(self) -> list[str]:
        return [smiles for smiles in self.states]
    

    def get_rdmol(self) -> list[Chem.Mol]:
        return [romol for smiles, (romol, charge) in self.states.items()]
    

    def get_pairs(self) -> list:
        return self.pairs
    


from rdworks import Mol, MolLibr
from rdworks.stereoisomers import enumerate_stereoisomers, enumerate_ring_bond_stereoisomers
from rdkit import Chem
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.EnumerateStereoisomers import EnumerateStereoisomers, StereoEnumerationOptions


def complete_stereoisomers(molecular_input: str | Chem.Mol | Mol, 
                           name: str | None = None, 
                           std: bool = False, 
                           override: bool = False, 
                           **kwargs) -> MolLibr:
    """Completes stereoisomers and returns a rdworks.MolLibr.

    Args:
        molecular_input (Union[Mol, str, Chem.Mol]): input molecule.
        name (Optional[str], optional): name of the molecule. Defaults to None.
        std (bool, optional): whether to standardize the input. Defaults to False.
        override (bool, optional): whether to override input stereoisomers. Defaults to False.

    Raises:
        TypeError: if `molecular_input` is not rdworks.Mol, SMILES, or rdkit.Chem.Mol object.

    Returns:
        MolLibr: a library of complete stereoisomers.
    """
    from rdworks import Mol, MolLibr
    
    if isinstance(molecular_input, Mol):
        if name:
            mol = molecular_input.rename(name)
        else:
            mol = molecular_input
    elif isinstance(molecular_input, str) or isinstance(molecular_input, Chem.Mol):
        mol = Mol(molecular_input, name, std)
    else:
        raise TypeError('complete_stereoisomers() expects rdworks.Mol, SMILES or rdkit.Chem.Mol object')
    
    ring_bond_stereo_info = mol.get_ring_bond_stereo()
    
    if override:
        mol = mol.remove_stereo()
    
    rdmols = enumerate_stereoisomers(mol.rdmol)
    # ring bond stereo is not properly enumerated
    # cis/trans information is lost if stereochemistry is removed,
    # which cannot be enumerated by EnumerateStereoisomers() function
    # so enumerate_ring_bond_stereoisomers() is introduced
    if len(ring_bond_stereo_info) > 0:
        ring_cis_trans = []
        for rdmol in rdmols:
            ring_cis_trans += enumerate_ring_bond_stereoisomers(rdmol,
                                                     ring_bond_stereo_info,
                                                     override=override)
        if len(ring_cis_trans) > 0:
            rdmols = ring_cis_trans
    
    if len(rdmols) > 1:
        libr = MolLibr(rdmols).unique().rename(mol.name, sep='.').compute(**kwargs)
    else:
        libr = MolLibr(rdmols).rename(mol.name).compute(**kwargs)
    
    for _ in libr:
        _.props.update(mol.props)
    
    return libr


def can_have_stereo(bond):
    """Check if a double bond can have E/Z stereochemistry"""
    atom1, atom2 = bond.GetBeginAtom(), bond.GetEndAtom()
    
    # Check if both atoms have at least 2 non-hydrogen neighbors
    for atom in [atom1, atom2]:
        non_h_neighbors = [n for n in atom.GetNeighbors() if n.GetAtomicNum() != 1]
        if len(non_h_neighbors) < 2:
            return False
    
    return True


def complete_tautomers(mol: Mol, **kwargs) -> MolLibr:
    """Returns a library of enumerated tautomers.

    Args:
        mol (Mol): input molecule.

    Returns:
        MolLibr: a library of enumerated tautomers.
    """
    enumerator = rdMolStandardize.TautomerEnumerator()
    enumerator.SetRemoveSp3Stereo(False)
    enumerator.SetRemoveBondStereo(False)  # Don't remove existing stereo
    tautomers = list(enumerator.Enumerate(mol.rdmol))
    
    all_stereoisomers = []
    
    for tautomer in tautomers:
        # Find unspecified double bonds that could have E/Z stereo
        unspecified_bonds = []
        for bond in tautomer.GetBonds():
            if (bond.GetBondType() == Chem.BondType.DOUBLE and 
                bond.GetStereo() == Chem.BondStereo.STEREONONE and
                can_have_stereo(bond)):
                unspecified_bonds.append(bond.GetIdx())
        
        if unspecified_bonds:
            # Enumerate all possible stereoisomers
            opts = StereoEnumerationOptions(onlyUnassigned=True, maxIsomers=50)
            stereoisomers = list(EnumerateStereoisomers(tautomer, options=opts))
            all_stereoisomers.extend(stereoisomers)
        else:
            all_stereoisomers.append(tautomer)
    
    if len(all_stereoisomers) > 1: 
        return MolLibr(all_stereoisomers).unique().rename(mol.name, sep='.').compute(**kwargs)
    
    return MolLibr(all_stereoisomers).compute(**kwargs)


from xml.etree.ElementTree import Element,SubElement,Comment,tostring,parse
from xml.dom import minidom
from rdkit import Chem

""" 
Jeroen Kazius, Ross McGuire, and Roberta Bursi, 2004. 
Derivation and Validation of Toxicophores for Mutagenicity Prediction. 
J. Med. Chem., 2005, 48 (1), pp 312–320 DOI: 10.1021/jm040835a
"""

count= 0
A= Element('Kazius2005')

for (name, smarts) in [
  ["Aromatic nitro","a[$([NX3](=O)=O),$([NX3+](=O)[O-])][!#8]"],
  ["Aromatic amine","a[NX3;H2]"],
  ["3-membered heterocycle","[N,O,S]1[C][C]1"],
  ["unsubstituted heteroatom bonded heteroatom","[NX2;H2,OX1;H1][N,O]"],
  ["Specific aromatic nitro","$(O=N(~O)a);!$(O=N(O)c[$(aS(=O)=O),$(aaS(=O)=O),$(aaaS(=O)=O),$(aC((F)F)F),$(aaC((F)F)F),$(aaaC((F)F)F)])"],
  ["Specific aromatic amine","a[NH2]"],
  ["aromatic nitroso","a[N;X2]=O"],
  ["alkyl nitrite","CO[N;X2]=O"],
  ["nitrosamine","N[N;X2]=O"],
  ["epoxide","O1[c,C]-[c,C]1"],
  ["aziridine","C1NC1"],
  ["azide","N=[N+]=[N-]"],
  ["diazo","C=[N+]=[N-]"],
  ["triazene","N=N-N"],
  ["unsubstituted heteroatom-bonded heteroatom","[OH,NH2][N,O]"],
  ["aromatic hydroxylamine","[OH]Na"],
  ["aliphatic halide","[Cl,Br,I]C"],
  ["carboxylic acid halide","[Cl,Br,I]C=O"],
  ["nitrogen or sulphur mustard","[N,S]!@[C;X4]!@[CH2][Cl,Br,I]"],
  ["bay-region in Polycyclic Aromatic Hydrocarbons","[cH]1[cH]ccc2c1c3c(cc2)cc[cH][cH]3"],
  ["K-region in Polycyclic Aromatic Hydrocarbons","[cH]1cccc2c1[cH][cH]c3c2ccc[cH]3"],
  ["sulphonate-bonded carbon (alkyl alkane sulphonate or dialkyl sulphate)","[$([C,c]OS(=O)(=O)O!@[c,C]),$([c,C]S(=O)(=O)O!@[c,C])]"],
  ["aliphatic N-nitro","O=N(~O)N"],
  ["aB unsaturated aldehyde (including a-carbonyl aldehyde)","[$(O=[CH]C=C),$(O=[CH]C=O)]"],
  ["diazonium","[N;v4]#N"],
  ["beta-propiolactone","O=C1CCO1"],
  ["ab unsaturated alkoxy group","[CH]=[CH]O"],
  ["1-aryl-2-monoalkyl hydrazine","[NH;!R][NH;R]a"],
  ["aromatic methylamine","[CH3][NH]a;![CH3][NH]a[$(a[$(C((F)F)F),$(S=O),$(C(=O)O)]),$(aa[$(C((F)F)F),$(S=O),$(C(=O)O)]),$(aaa[$(C((F)F)F),$(S=O),$(C(=O)O)])]"],
  ["ester derivative of aromatic hydroxylamine (including original specific toxicophore)","aN([$([OH]),$(O*=O)])[$([#1]),$(C(=O)[CH3]),$([CH3]),$([OH]),$(O*=O)]"],
  ["polycyclic planar system 1","a13~a~a~a~a2~a1~a(~a~a~a~3)~a~a~a~2"],
  ["polycyclic planar system 2","a1~a~a~a2~a~1~a~a3~a(~a~2)~a~a~a~3"],
  ["polycyclic planar system 3","a1~a~a~a2~a~1~a~a~a3~a~2~a~a~a~3"],
  ["polycyclic planar system 4","a1~a~a~a~a2~a~1~a3~a(~a~2)~a~a~a~a~3"],
  ["polycyclic planar system 5","a1~a~a~a~a2~a~1~a~a3~a(~a~2)~a~a~a~3"],
  ["polycyclic planar system 6","a1~a~a~a~a2~a~1~a~a3~a(~a~2)~a~a~a~a~3"],
  ["polycyclic planar system 7","a1~a~a~a~a2~a~1~a~a~a3~a~2~a~a~a~3"],
  ["polycyclic planar system 8","a1~a~a~a~a2~a~1~a~a~a3~a~2~a~a~a~a~3"],
  ["polycyclic planar system 9","a13~a~a~a~a2~a1~a(~a~a~a~3)~a~a~2"],
     ]:
     m = Chem.MolFromSmarts(smarts)
     if m is None: continue
     count += 1
     entry= SubElement (A, 'group')
     entry.set('name', '('+str(count)+') '+name)
     node= SubElement(entry, "SMARTS")
     node.text= smarts
   
g=open('Kazius2005.xml','wt')
coarse= tostring(A,'utf-8')
g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )
g.close()



from pathlib import Path

from rdkit import Chem
from rdkit.Chem import AllChem, rdmolfiles

from rdworks import Conf, Mol, MolLibr
from rdworks.utils import compute, precheck_path, guess_mol_id

import pandas as pd
import re
import gzip


conf_name_convention = re.compile(r'[a-zA-Z0-9-_.!@#$%^&*()+=]+.[0-9]+/[0-9]+')


def read_csv(path:str | Path, smiles:str, name:str, std:bool=False, **kwargs) -> MolLibr:
    """Returns a library of molecules reading from a .csv file.

    Other columns will be read as properties.

    Args:
        path (Union[str, Path]): filename or path to a .csv file.
        smiles (str): column for SMILES.
        name (str): column for name.
        std (bool, optional): whether to standardize the input. Defaults to False.

    Raises:
        ValueError: if `smiles` or `name` column is not found in the csv file.

    Returns:
        MolLibr: a library of molecules.
    """
    path = precheck_path(path)
    df = pd.read_csv(path)
    try:
        assert smiles in list(df.columns)
    except:
        raise ValueError(f"Cannot find SMILES column (`smiles=`) {smiles}")
    try:
        assert name in list(df.columns)
    except:
        raise ValueError(f"Cannot find NAME column (`name=`) {name}")
    
    largs = [ (smiles, name, std) for smiles, name in zip(list(df[smiles]), list(df[name])) ]
    libr = MolLibr(compute(Mol, largs, desc='Reading CSV', **kwargs))

    # read other columns as properties
    # A list of dictionaries, where each dictionary represents a row, 
    # with column names as keys and cell values as values: 
    # [{column -> value}, ..., {column -> value}].
    csv_records = df.to_dict('records')
    for mol, row_dict in zip(libr, csv_records):
        mol.props.update({ k:v for (k,v) in row_dict.items() if k not in [smiles, name]})

    return libr


def merge_csv(libr: MolLibr, path:str | Path, on:str='name') -> MolLibr:
    """Returns a copy of MolLibr merged with properties from `on` column of a .csv file.

    Args:
        libr (MolLibr): library to be merged.
        path (Union[str, Path]): filename or path to a .csv file.
        on (str, optional): column for name. Defaults to 'name'.

    Raises:
        ValueError: if `on` column is not found in the csv file.

    Returns:
        MolLibr: a copy of library of molecules.
    """
    path = precheck_path(path)
    df = pd.read_csv(path)
    try:
        assert on in list(df.columns)
    except:
        raise ValueError(f"Cannot find ON column (`on=`) {on}")
    # A list of dictionaries, where each dictionary represents a row, 
    # with column names as keys and cell values as values: 
    # [{column -> value}, ..., {column -> value}].
    csv_records = df.to_dict('records')
    data = {}
    for row_dict in csv_records:
        data[row_dict[on]] = { k:v for (k,v) in row_dict.items() if k != on }
    
    merged_libr = libr.copy()

    for mol in merged_libr:
        if mol.name in data: # mol.props can be partly updated from csv
            mol.props.update(data[mol.name])
    
    return merged_libr


def read_dataframe(df:pd.DataFrame, smiles:str, name:str, std:bool=False) -> MolLibr:
    """Returns rdworks.MolLibr object from a pandas DataFrame.

    Args:
        df (pd.DataFrame): pandas.DataFrame.
        smiles (str): column for SMILES.
        name (str): column for name.
        std (bool, optional): whether to standardize the input. Defaults to False.

    Raises:
        TypeError: if `df` is not pandas DataFrame.
        ValueError: if `smiles` or `name` column is not found.

    Returns:
        MolLibr: a library of molecules.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"Expects a pandas.DataFrame object")
    try:
        assert smiles in list(df.columns)
    except:
        raise ValueError(f"Cannot find SMILES column (`smiles=`) {smiles}")
    try:
        assert name in list(df.columns)
    except:
        raise ValueError(f"Cannot find NAME column (`name=`) {name}")
    
    return MolLibr(list(df[smiles]), list(df[name]), std=std)



def read_smi(path:str | Path, std:bool = False, **kwargs) -> MolLibr:
    """Read a SMILES file and create a molecular library.

    Args:
        path (str | Path): path to the SMILES file.
        std (bool, optional): whether to standardize. Defaults to False.

    Raises:
        FileNotFoundError: when path does not exist.

    Returns:
        MolLibr: a library of molecules.
    """
    if not Path(path).exists():
        raise FileNotFoundError(f"Path {path} does not exist.")
    if path.suffix == '.gz':
        with gzip.open(path, "rb") as gz:
            largs = [ tuple(line.decode('utf-8').strip().split()[:2] + [std]) for line in gz ]
    else:
        with open(path, "r") as smi:
            largs = [ tuple(line.strip().split()[:2] +[std]) for line in smi ]
    return MolLibr(compute(Mol, largs, desc='Reading SMILES', **kwargs))



def _map_sdf(rdmol:Chem.Mol, name:str, std:bool, props:dict) -> Mol:
    """A map function for `read_sdf()` to return a rdworks.Mol object reading from a SDF entry.

    Args:
        rdmol (Chem.Mol): input molecule.
        name (str): name of the molecule.
        std (bool): whether to standardize the input SMILES.
        props (dict): dictionary of molecule properties.

    Returns:
        Mol: rdworks.Mol object.
    """
    obj = Mol(rdmol, name, std)
    obj.props = props
    return obj


def read_sdf(path:str | Path, std:bool=False, confs:bool=False, props:bool=True, **kwargs) -> MolLibr:
    """Returns a library of molecules reading from a SDF file.

    Args:
        path (Union[str, PosixPath]): filename or path to the .sdf file.
        std (bool, optional): whether to standardize the input. Defaults to False.
        confs (bool, optional): whether to read 3D conformers and keep hydrogens. Defaults to False.
        props (bool, optional): whether to read SDF properties. Defaults to True.

    Returns:
        MolLibr: a library of molecules.
    """
    path = precheck_path(path)
    if path.suffix == '.gz':
        with gzip.open(path, 'rb') as gz:
            # switch ^ True, XOR(^) inverts only if switch is True
            with Chem.ForwardSDMolSupplier(gz, sanitize=True, removeHs=(confs ^ True)) as gzsdf:
                    lrdmols = [ m for m in gzsdf if m is not None ]
    else:
        # switch ^ True, XOR(^) inverts only if switch is True
        with Chem.SDMolSupplier(path, sanitize=True, removeHs=(confs ^ True)) as sdf:
            lrdmols = [ m for m in sdf if m is not None ]
    
    if props:
        lprops = [ m.GetPropsAsDict() for m in lrdmols ]
        try:
            lnames = [ m.GetProp('_Name') for m in lrdmols ]
            assert len(set(lnames)) == len(lrdmols)
        except:
            (k, c, t) = guess_mol_id(lprops)
            if k is None:
                for i, m in enumerate(lrdmols, start=1):
                    name = f'_{i}_'
                    lnames.append(name)
            else:
                lnames = []
                for i, m in enumerate(lrdmols, start=1):
                    try:
                        name = m.GetProp(k)
                    except:
                        name = f'_{i}_'
                    lnames.append(name)
    else:
        lprops = [ None ] * len(lrdmols)
        lnames = [ None ] * len(lrdmols)
    
    largs = [ (rdmol, name, std, props) for rdmol, name, props in zip(lrdmols, lnames, lprops) ]

    obj = MolLibr()
    if confs: 
        # reading 3D SDF (conformers)
        last_smiles = None
        new_mol = None
        for rdmol, name, props in zip(lrdmols, lnames, lprops):
            # rdworks name convention (e.g. xxxx.yy/zzz)
            if conf_name_convention.match(name):
                (isomer_name, _) = name.split('/')
            else:
                isomer_name = name
            smiles = Chem.MolToSmiles(rdmol)  # canonicalized SMILES
            if last_smiles is None or last_smiles != smiles:
                if new_mol:
                    obj.libr.append(new_mol.rename())
                # start a new molecule
                rdmol_2d = Chem.RemoveHs(rdmol)
                AllChem.Compute2DCoords(rdmol_2d)
                # initialize a new molecule with the H-removed 2D
                new_mol = Mol(rdmol_2d, isomer_name, std=False) # atom indices remain unchanged.
            new_conf = Conf(rdmol)
            new_conf.props.update(props)
            new_mol.confs.append(new_conf)
            last_smiles = smiles
        if new_mol: # handle the last molecule
            obj.libr.append(new_mol.rename())
    else: 
        # reading 2D SDF
        obj = MolLibr(compute(_map_sdf, largs, desc='Reading SDF', **kwargs))

    return obj



def read_mae(path:str | Path, std:bool=False, confs:bool=True, **kwargs) -> MolLibr:
    """Returns a library of molecules reading from a Schrodinger Maestro file.

    Args:
        path (Union[str, Path]): filename or path to the .mae or .maegz file.
        std (bool, optional): whether to standardize the input. Defaults to False.
        confs (bool, optional): whether to read 3D conformers. Defaults to True.

    Returns:
        MolLibr: a library of molecules.
    """
    path = precheck_path(path)

    if path.suffix == '.maegz':
        with gzip.open(path, 'rb') as gz:
            # switch ^ True, XOR(^) inverts only if switch is True
            with rdmolfiles.MaeMolSupplier(gz, sanitize=True, removeHs=(confs ^ True)) as maegz:
                    lrdmols = [ m for m in maegz if m is not None ]
    else:
        # switch ^ True, XOR(^) inverts only if switch is True
        with rdmolfiles.MaeMolSupplier(path, sanitize=True, removeHs=(confs ^ True)) as mae:
            lrdmols = [ m for m in mae if m is not None ]

    lnames = [m.GetProp('_Name') for m in lrdmols]
    largs = [(rdmol, name, std) for rdmol, name in zip(lrdmols, lnames)]
    
    obj = MolLibr()

    if confs: # reading 3D SDF (conformers)
        last_smiles = None
        new_mol = None
        for rdmol, name in zip(lrdmols, lnames):
            # rdworks name convention (e.g. xxxx.yy/zzz)
            if conf_name_convention.match(name):
                (isomer_name, _) = name.split('/')
            else:
                isomer_name = name
            smiles = Chem.MolToSmiles(rdmol)  # canonicalized SMILES
            if last_smiles is None or last_smiles != smiles:
                if new_mol:
                    obj.libr.append(new_mol.rename())
                # start a new molecule
                # !!!! rdmol and new_mol do not have consistent atom indices !!!
                # idxmap: original atom index -> canonicalized rdmol atom index
                # smiles = Chem.MolToSmiles(rdmol) # canonicalization creates `_smilesAtomOutputOrder` property
                # idxord_o = ast.literal_eval(rdmol.GetProp("_smilesAtomOutputOrder"))
                # idxmap_o = {o.GetIdx():idxord_o.index(o.GetIdx()) for o in rdmol.GetAtoms()}
                rdmol_2d = Chem.RemoveHs(rdmol)
                AllChem.Compute2DCoords(rdmol_2d)
                new_mol = Mol(rdmol_2d, isomer_name, std=False) # atom indices remain unchanged.

            new_mol.confs.append(Conf(rdmol))
            
            last_smiles = smiles
        if new_mol: # handle the last molecule
            obj.libr.append(new_mol.rename())

    else: # reading 2D SDF
        obj = MolLibr(compute(Mol, largs, desc='Reading Mae', **kwargs))
    
    return obj


import rdworks
from pathlib import Path

datadir = Path(__file__).parent.resolve() / "data"

def test_readin_qupkake_output():
    libr = rdworks.read_sdf(datadir / 'qupkake_output_SAMPL6_SM07.sdf', confs=True)
    # libr contains one molecule because of confs=True
    assert libr.count() == 1
    for m in libr:
        assert m.count() == 3
        print(m.props)
        print()
        for c in m:
            print(c.props)
            print()

        print(m.serialize(compressed=True))


if __name__ == '__main__':
    test_readin_qupkake_output()



from rdworks import Mol
from rdworks.conf import prep_confs_batches
from rdkit import Chem


class Dummy_BatchSingplePoint():
    def __init__(self, rdmols):
        self.rdmols = rdmols

    def __str__(self):
        return "Dummy_BatchSinglePoint"

    def run(self, **kwargs):
        from collections import namedtuple
        Calculated = namedtuple('Calculated', ['mols',])
        for rdmol in self.rdmols:
            rdmol.SetProp('E_tot(kcal/mol)', '1.0')
        return Calculated(mols=self.rdmols)


class Dummy_BatchOptimizer():
    def __init__(self, rdmols):
        self.rdmols = rdmols

    def __str__(self):
        return "Dummy_BatchOptimizer"

    def run(self, **kwargs):
        from collections import namedtuple
        Optimized = namedtuple('Optimized', ['mols',])
        for rdmol in self.rdmols:
            rdmol.SetProp('E_tot_init(kcal/mol)', '10.0')
            rdmol.SetProp('E_tot(kcal/mol)', '1.0')
            rdmol.SetProp('Converged', 'True')
        return Optimized(mols=self.rdmols)


def test_prep_confs_batches():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs(n=50, method='ETKDG')
    batches = prep_confs_batches(mol.confs)
    assert isinstance(batches, list)
    assert len(batches) > 0
    assert isinstance(batches[0].rdmols, list)
    assert isinstance(batches[0].rdmols[0], Chem.Mol)
    assert isinstance(batches[0].size, int)
    assert batches[0].size > 0
    assert isinstance(batches[0].num_atoms, int)
    assert batches[0].num_atoms > 0


def test_singlepoint_batch():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs(n=10, method='ETKDG')
    mol = mol.singlepoint(calculator=Dummy_BatchSingplePoint, batchsize_atoms=16384)
    assert all([_.props.get('E_tot(kcal/mol)') is not None for _ in mol.confs])


def test_optimize_confs_batch():
    mol = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer")
    mol = mol.make_confs(n=50, method='ETKDG')
    mol = mol.optimize_confs(calculator=Dummy_BatchOptimizer, batchsize_atoms=16384)
    assert all([_.props.get('E_tot_init(kcal/mol)') is not None for _ in mol.confs])
    assert all([_.props.get('E_tot(kcal/mol)') is not None for _ in mol.confs])
    assert all([_.props.get('Converged') is not None for _ in mol.confs])


import numpy as np
import math
import operator
import gzip
import base64

from pathlib import Path
from types import SimpleNamespace
from typing import Any
from collections.abc import Callable
from concurrent.futures import ProcessPoolExecutor
from tqdm import tqdm

from rdkit import Chem

from rdworks.autograph.centroid import centroid_medoid


def compress_string(text: str) -> str:
    """compress string to base64-encoded string.

    Args:
        text (str): original string.

    Returns:
        str: base64-encoded compressed string.
    """
    compressed_bytes = gzip.compress(text.encode('utf-8'))
    return base64.b64encode(compressed_bytes).decode('utf-8')


def decompress_string(compressed_text: str) -> str:
    """decompress base64-encoded string to original string.

    Args:
        compressed_text (str): base64-encoded compressed string.

    Returns:
        str: original string.
    """
    compressed_bytes = base64.b64decode(compressed_text)
    decompressed_text = gzip.decompress(compressed_bytes).decode('utf-8')
    return decompressed_text


def compute(fn:Callable, largs: list, **kwargs) -> list:
    max_workers = kwargs.get('max_workers', 1)
    chunksize   = kwargs.get('chunksize', 10)
    progress    = kwargs.get('progress', False)
    desc = kwargs.get('desc', 'Progress')
    n = len(largs)
    if max_workers > 1:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            if progress:
                results = list(tqdm(executor.map(fn, largs, chunksize=chunksize), desc=desc, total=n))
            else:
                results = list(executor.map(fn, largs, chunksize=chunksize))
    else:
        if progress:
            results = [ fn(*larg) for larg in tqdm(largs, desc=desc, total=n) ]
        else:
            results = [ fn(*larg) for larg in largs ]
    return results



def precheck_path(path:str | Path) -> Path:
    """Prechecks filename or path and returns a string for the pathlib.PosixPath.

    Args:
        path (Union[str, PosixPath]): filename or path.
        origin (str): origin of data.

    Raises:
        FileNotFoundError: if the path is not found.

    Returns:
        str: a string for the path.
    """
    if isinstance(path, Path):
        pass
    elif isinstance(path, str):
        path = Path(path)
    if path.exists() and path.is_file():
        return path
    else:
        raise FileNotFoundError(f"File path {path.as_posix()} does not exist.")
    


def guess_mol_id(lprops:list[dict]) -> tuple[str, int, int]:
    """Guesses molecular ID from SDF properties.

    Molecular ID is guessed by the coverage(=count of unique values divided by total count).
    A property suitable for ID should have coverage of 1.0.

    Args:
        lprops (List[dict]): a list of properties.

    Returns:
        Tuple[str, int, int]: (property, count of unique values, total count)
    """
    f = {} # unique occurrence
    for props in lprops:
        for k in props:
            v = props[k]
            # float is not suitable for molecular id
            if isinstance(v, float): 
                continue 
            if not (k in f): 
                f[k] = set()
            # str(int) is acceptable for molecular id
            if isinstance(v, int): 
                f[k].add(str(v))
            elif isinstance(v, str):
                f[k].add(v)
    r = [(k, len(f[k]), -max([len(x) for x in f[k]])) for k in f]
    r = sorted(r, key=operator.itemgetter(1,2))
    try:
        (property_key, count, total) = (r[-1][0], r[-1][1], len(lprops))
    except IndexError:
        (property_key, count, total) = (None, 0, 0)
    return (property_key, count, total)


def dict_to_simplenamespace(data):
    if isinstance(data, dict):
        return SimpleNamespace(**{k: dict_to_simplenamespace(v) for k, v in data.items()})
    elif isinstance(data, list):
        return [dict_to_simplenamespace(item) for item in data]
    else:
        return data
    

def recursive_round(data:Any, decimals:int=2) -> Any:
    """Recursively round float values to a given decimal places.

    Args:
    data: The input data, which can be a list, dictionary, or any
            other data type. It can contain nested lists and dictionaries.
    decimals: number of decimal places.
    """
    if not isinstance(decimals, int) or decimals < 0:
        raise ValueError("decimals must be a non-negative integer.")

    def _recursive_round(current_item):
        if isinstance(current_item, float):
            return round(current_item, decimals)
        elif isinstance(current_item, np.float64):
            return round(float(current_item), decimals)
        elif isinstance(current_item, list):
            return [_recursive_round(item) for item in current_item]
        elif isinstance(current_item, dict):
            return {key: _recursive_round(value) for key, value in current_item.items()}
        else:
            return current_item

    return _recursive_round(data)


def fix_decimals_in_list(in_list:list, decimals:int=2) -> list:
    """Fixes the decimal places of all float values in a list.

    Args:
        list: The list to fix.
        decimals (int): The number of decimal places to fix the float values to.

    Returns:
        list: a list with the float values fixed to the specified number of decimal places.
    """

    out_list = []
    for item in in_list:
        if isinstance(item, float):
            out_list.append(round(item, decimals))
        elif isinstance(item, dict):
            out_list.append(fix_decimals_in_dict(item, decimals))
        elif isinstance(item, list) or isinstance(item, tuple):
            out_list.append(fix_decimals_in_list(item, decimals))
        else:
            out_list.append(item)
    return out_list


def fix_decimals_in_dict(in_dict:dict, decimals:int=2) -> dict:
    """Fixes the decimal places of all float values in a dictionary.

    Args:
        dictionary: The dictionary to fix.
        decimals (int): The number of decimal places to fix the float values to.

    Returns:
        dict: a dictionary with the float values fixed to the specified number of decimal places.
    """
    out_dict = {}
    for k, v in in_dict.items():
        if isinstance(v, float):
            out_dict[k] = round(v, decimals)
        elif isinstance(v, list) or isinstance(v, tuple):
            out_dict[k] = fix_decimals_in_list(v, decimals)
        elif isinstance(v, dict):
            out_dict[k] = fix_decimals_in_dict(v, decimals)
        else:
            out_dict[k] = v
    return out_dict


def convert_tril_to_symm(lower_triangle_values:list) -> np.ndarray:
    """Converts lower triangle values to a symmetric full matrix.

    Args:
        lower_triangle_values (list): list of lower triangle matrix values.

    Returns:
        np.ndarray: numpy array of a symmetric full matrix.
    """
    n = math.ceil(math.sqrt(len(lower_triangle_values)*2))
    rmsd_matrix = np.zeros((n,n))
    rmsd_matrix[np.tril_indices(n, k=-1)] = lower_triangle_values
    symm_matrix = np.maximum(rmsd_matrix, rmsd_matrix.transpose())
    return symm_matrix


def convert_triu_to_symm(upper_triangle_values:list) -> np.ndarray:
    """Converts upper triangle values to a symmetric full matrix.

    Args:
        upper_triangle_values (list): list of upper triangle matrix values.

    Returns:
        np.ndarray: numpy array of a symmetric full matrix.
    """
    n = math.ceil(math.sqrt(len(upper_triangle_values)*2))
    rmsd_matrix = np.zeros((n,n))
    rmsd_matrix[np.triu_indices(n, k=1)] = upper_triangle_values
    symm_matrix = np.maximum(rmsd_matrix, rmsd_matrix.transpose())
    return symm_matrix


def _QT_diameter(rmsd_matrix:np.ndarray, A:list) -> float:
    """A subroutine for `QT()` to returns the maximum pairwise distance.

    Args:
        rmsd_matrix (np.ndarray): numpy array of rmsd.
        A (list): list of indexes.

    Returns:
        float: maximum pairwise distance.
    """
    return np.max(rmsd_matrix[A][:,A])


def _QT_clustering(rmsd_matrix:np.ndarray, G:set, threshold:float, clusters:list) -> list:
    """A subroutine for `QT()` to perform QTC algorithm.

    Args:
        rmsd_matrix (np.ndarray): pairwise rmsd matrix.
        G (set): set of indexes used for recursive calling.
        threshold (float): quality threshold (A).
        clusters (list): list of clusters used for recursive calling.

    Returns:
        list: a list of final clusters.
    """

    if len(G) <= 1:
        clusters.append(G)
        return
    C = [] # cluster candidates
    for i in G:
        flag = True
        A = [i]
        A_diameter = 0.0 # max of pairwise distances
        while flag and A != G:
            # find j that minimize diameter of A + [j]
            diameters = [(_QT_diameter(rmsd_matrix, A + [j]), j) for j in G if j not in A]
            if len(diameters) == 0:
                flag = False
            else:
                (min_diameter, min_j) = min(diameters, key=lambda x: x[0])
                if min_diameter > threshold:
                    flag = False
                else:
                    A += [min_j]
                    A_diameter = min_diameter
        C.append((A, A_diameter))
    C = sorted(C, key=lambda x: (len(x[0]), -x[1]), reverse=True)
    # if cardinality of C is tied, smaller diameter is picked
    largest_C = set(C[0][0])
    clusters.append(largest_C)
    _QT_clustering(rmsd_matrix, G-largest_C, threshold, clusters)


def QT(rmsd_matrix:np.ndarray, threshold:float) -> tuple:
    """Perform QT clustering.

    Args:
        rmsd_matrix (np.ndarray): pairwise rmsd matrix.
        threshold (float): quality threshold (A)

    Returns:
        tuple: (cluster assignment, centroid indices)
    """
    N = rmsd_matrix.shape[0]
    clusters = []
    _QT_clustering(rmsd_matrix, set(list(range(N))), threshold, clusters)
    # ex. clusters=  [{6, 7, 11}, {4, 5, 8}, {0}, {1}, {10}, {9}, {2}, {3}]
    cluster_assignment = [None,] * N
    for cluster_idx, indices in enumerate(clusters):
        for conf_idx in indices:
            cluster_assignment[conf_idx] = cluster_idx
    centroid_indices = centroid_medoid(cluster_assignment, rmsd_matrix)
    return cluster_assignment, centroid_indices


# def rdmol_to_graph(rdmol:Chem.Mol) -> nx.Graph:
#     """Converts rdkit.Chem.Mol to a networkx graph object.

#     Args:
#         rdmol (Chem.Mol): input molecule.

#     Returns:
#         nx.Graph: networkx graph object.
#     """
#     G = nx.Graph()
#     for atom in rdmol.GetAtoms():
#         G.add_node(atom.GetIdx(), # 0-based index
#                    atomic_num=atom.GetAtomicNum(),
#                    formal_charge=atom.GetFormalCharge(),
#                    chiral_tag=atom.GetChiralTag(),
#                    hybridization=atom.GetHybridization(),
#                    num_explicit_hs=atom.GetNumExplicitHs(),
#                    is_aromatic=atom.GetIsAromatic())
#     for bond in rdmol.GetBonds():
#         G.add_edge(bond.GetBeginAtomIdx(),
#                    bond.GetEndAtomIdx(),
#                    bond_type=bond.GetBondType())
#     return G


# def rdmol_to_graph_(rdmol:Chem.Mol) -> nx.Graph:
#     """Converts rdkit.Chem.Mol to a networkx graph object (another implementation).

#     Args:
#         rdmol (Chem.Mol): input molecule.

#     Returns:
#         nx.Graph: networkx graph object.
#     """ 
#     atomic_nums = [atom.GetAtomicNum() for atom in rdmol.GetAtoms()]
#     formal_charges = [atom.GetFormalCharge() for atom in rdmol.GetAtoms()]
#     ad_matrix = Chem.GetAdjacencyMatrix(rdmol, useBO=True)
#     # useBO: (optional) toggles use of bond orders in calculating the matrix. Default value is 0.
#     # RETURNS: a Numeric array of floats containing the adjacency matrix
#     # [[0. 1. 0. 0. 0. 0. 0. 0. 0.]
#     # [1. 0. 1. 1. 1. 0. 0. 0. 0.]
#     # [0. 1. 0. 0. 0. 0. 0. 0. 0.]
#     # [0. 1. 0. 0. 0. 0. 0. 0. 0.]
#     # [0. 1. 0. 0. 0. 1. 0. 1. 0.]
#     # [0. 0. 0. 0. 1. 0. 2. 0. 0.]
#     # [0. 0. 0. 0. 0. 2. 0. 0. 0.]
#     # [0. 0. 0. 0. 1. 0. 0. 0. 2.]
#     # [0. 0. 0. 0. 0. 0. 0. 2. 0.]]
#     for i,(a_num,f_c) in enumerate(zip(atomic_nums, formal_charges)):
#         if f_c !=0:
#             ad_matrix[i,i] = a_num + f_c
#         else:
#             ad_matrix[i,i] = a_num
#     G = nx.from_numpy_array(ad_matrix)
#     return G


# def graph_to_rdmol(G:nx.Graph) -> Chem.Mol:
#     """Converts a networkx graph object to rdkit.Chem.Mol object.

#     Args:
#         G (nx.Graph): a networkx graph.

#     Returns:
#         Chem.Mol: rdkit.Chem.Mol object.
#     """
#     rdmol = Chem.RWMol()
#     atomic_nums = nx.get_node_attributes(G, 'atomic_num')
#     chiral_tags = nx.get_node_attributes(G, 'chiral_tag')
#     formal_charges = nx.get_node_attributes(G, 'formal_charge')
#     node_is_aromatics = nx.get_node_attributes(G, 'is_aromatic')
#     node_hybridizations = nx.get_node_attributes(G, 'hybridization')
#     num_explicit_hss = nx.get_node_attributes(G, 'num_explicit_hs')
#     node_to_idx = {}
#     for node in G.nodes():
#         a=Chem.Atom(atomic_nums[node])
#         a.SetChiralTag(chiral_tags[node])
#         a.SetFormalCharge(formal_charges[node])
#         a.SetIsAromatic(node_is_aromatics[node])
#         a.SetHybridization(node_hybridizations[node])
#         a.SetNumExplicitHs(num_explicit_hss[node])
#         idx = rdmol.AddAtom(a)
#         node_to_idx[node] = idx
#     bond_types = nx.get_edge_attributes(G, 'bond_type')
#     for edge in G.edges():
#         first, second = edge
#         ifirst = node_to_idx[first]
#         isecond = node_to_idx[second]
#         bond_type = bond_types[first, second]
#         rdmol.AddBond(ifirst, isecond, bond_type)
#     Chem.SanitizeMol(rdmol)
#     return rdmol


__version__ = '0.59.1'

from rdworks.conf    import Conf
from rdworks.mol     import Mol
from rdworks.mollibr import MolLibr

from rdworks.readin import read_csv, merge_csv, read_dataframe, read_smi, read_sdf, read_mae 
from rdworks.std import desalt_smiles, standardize_smiles, standardize

from rdworks.complete import complete_stereoisomers, complete_tautomers
from rdworks.ionized import IonizedStates

from rdworks.rgroup import expand_rgroup, most_common, most_common_in_NP
from rdworks.scaffold import scaffold_network, scaffold_tree, BRICS_fragmented, BRICS_fragment_indices
from rdworks.matchedseries import MatchedSeries

from rdworks.descriptor import rd_descriptor, rd_descriptor_f
from rdworks.xml import list_predefined_xml, get_predefined_xml, parse_xml

import rdkit
import logging

__rdkit_version__ = rdkit.rdBase.rdkitVersion

rdkit_logger = rdkit.RDLogger.logger().setLevel(rdkit.RDLogger.CRITICAL)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO) # level: DEBUG < INFO < WARNING < ERROR < CRITICAL

logger_stream = logging.StreamHandler() # sys.stdout or sys.stderr
logger_format = logging.Formatter(fmt='%(asctime)s %(levelname)s %(message)s', 
                                  datefmt='%Y-%m-%d %H:%M:%S')
logger_stream.setFormatter(logger_format)
logger.addHandler(logger_stream)
